package org.apache.ojb.odmg;

/* Copyright 2002-2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import javax.transaction.Status;
import javax.transaction.Synchronization;

import org.apache.ojb.broker.OJBRuntimeException;
import org.apache.ojb.broker.util.logging.Logger;
import org.apache.ojb.broker.util.logging.LoggerFactory;

/**
 * @author <a href="mailto:mattbaird@yahoo.com">Matthew Baird</a>
 *
 * @version $Id: J2EETransactionImpl.java,v 1.25 2004/05/03 23:05:56 arminw Exp $
 */
public class J2EETransactionImpl extends TransactionImpl implements Synchronization
{
    private Logger log = LoggerFactory.getLogger(J2EETransactionImpl.class);
    private boolean isInExternTransaction;

    /**
     * beforeCompletion is being called twice in JBoss, so this
     * isPrepared flag prevents code from executing twice.
     * todo: find out why it's being called twice and fix it.
     */
    private boolean m_isPrepared = false;

    public J2EETransactionImpl(DatabaseImpl theCurrentDB)
    {
        super(theCurrentDB);
        isInExternTransaction = false;
    }

    public void setInExternTransaction(boolean mode)
    {
        isInExternTransaction = mode;
    }

    public boolean isInExternTransaction()
    {
        return isInExternTransaction;
    }

    public void join()
    {
        throw new UnsupportedOperationException("Not supported in managed enviroment");
    }

    public void leave()
    {
        throw new UnsupportedOperationException("Not supported in managed enviroment");
    }

    public void checkpoint()
    {
        throw new UnsupportedOperationException("Not supported in managed enviroment");
    }

    /**
     * FOR internal use. This method was called after the external transaction was completed.
     * @see javax.transaction.Synchronization
     */
    public void afterCompletion(int status)
    {
        // no real transaction control, just cleanup.
        switch (status)
        {
            case Status.STATUS_COMMITTED:
                break;
            case Status.STATUS_ROLLEDBACK:
                break;
            default:
        }
    }

    /**
     * FOR internal use. This method was called before the external transaction was completed.
     * @see javax.transaction.Synchronization
     */
    public void beforeCompletion()
    {
        // avoid redundant calls
        if (m_isPrepared) return;
        OJBRuntimeException ex = null;
        try
        {
            try
            {
                JTATxManager mgr = (JTATxManager)TxManagerFactory.instance();
                if (mgr.getJTATransaction().getStatus() == Status.STATUS_MARKED_ROLLBACK)
                {
                    ex = new OJBRuntimeException("Could not commit, because tx status was marked rollback");
                }
                else
                {
                    try
                    {
                        prepare();
                        commit();
                        /*
                         MBAIRD: hack to get around the lack of swizzling.
                         arminw:
                         change the cache implementation to avoid swizzling
                         */
                        // getBroker().clearCache();
                    }
                    catch (Exception e)
                    {
                        ex = new OJBRuntimeException("beforeCompletion: commit failed", e);
                    }
                }
            }
            catch (Exception e)
            {
                ex = new OJBRuntimeException("beforeCompletion: commit failed", e);
            }
            if(ex != null)
            {
                log.error(ex);
                abort();
            }
        }
        finally
        {
            m_isPrepared = true;
            setInExternTransaction(false);
        }
    }

    public void commit()
    {
        try
        {
            // prepareForCommit was done before
            if (log.isDebugEnabled()) log.debug("Commit transaction " + this + ", commit on broker " + broker);
            getBroker().commitTransaction();
            doClose();
            setStatus(Status.STATUS_COMMITTED);
        }
        catch (Exception ex)
        {
            // We should not reach this block
            log.error("Error while commit used PB-Instance", ex);
            if (log.isDebugEnabled())
            {
                log.debug(ex.getMessage());
                ex.printStackTrace();
            }
            abort();
        }
    }

    public void abort()
    {
        try
        {
            doAbort();
            TxManagerFactory.instance().abortExternalTx(this);
            doClose();
            setStatus(Status.STATUS_ROLLEDBACK);
        }
        catch (Throwable t)
        {
            if (log.isDebugEnabled())
            {
                log.debug(t.getMessage());
                t.printStackTrace();
            }
            throw new OJBRuntimeException("Unexpected error occured while aborting", t);

        }
        finally
        {
            setInExternTransaction(false);
        }
        //Hack, throw this Exception to notify client of abort, without changing API
        throw new TransactionAbortedExceptionOJB("Internal ODMG transaction was aborted");
    }
}
