package org.apache.ojb.odmg.locking;

/* Copyright 2004-2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.ojb.broker.Identity;
import org.apache.ojb.broker.PersistenceBrokerException;


/**
 * @author Thomas Mahler
 */
public class LockServerServlet extends HttpServlet
{
	private static int numRequests;
	
	private static InMemoryLockMapImpl lockmap = new InMemoryLockMapImpl();
	
	private static Throwable lastError = null;
	
	

    /* (non-Javadoc)
     * @see javax.servlet.http.HttpServlet#doGet(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
     */
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
        throws ServletException, IOException
    {
		 response.setContentType("text/html");
		 response.setHeader("Pragma", "no-cache");

		 PrintWriter out = response.getWriter();

		 out.println(
			 "<html><head><title>OJB Distributed Locking Servlet Status Page</title>");
		 out.println("</head><body><h1>OJB Distributed Locking Servlet</h1>");
		 out.println("The servlet is running.<p>");

		 if (lastError == null) {
			 out.println("The LockServer is running.<p>");
			 out.println("Persistent Locks: " + lockmap.getSize() + "<p>");
			 out.println("Processed Lock Request: " + numRequests + "<p>");
		 } else {
			 out.println("<h2>The LockServer has a problem!</h2>");
			 out.println("The error message is:<p>");
			 out.println(lastError.getMessage() + "<p>");
			 lastError.printStackTrace(out);
			 lastError = null;
		 }

		 out.println("</body></html>");

    }

    /* (non-Javadoc)
     * @see javax.servlet.http.HttpServlet#doPost(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
     */
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
        throws ServletException, IOException
    {
    	// update counter
    	numRequests++;
    	    	
        try
        {
            // read request:
            byte[] req = getRequestByteArray(request);
            // first byte contains command selector 
            byte selector = req[0];
            // rest is data
            byte[] data = new byte[req.length -1];
            System.arraycopy(req,1,data,0,data.length);
            
            Object result = null;
            
            // now execute the command specified by the selector
            switch (selector)
            {
            	case 'w' :
            	{
            		result = getWriter(data);
            		break;
            	}
            	case 'r' :
            	{
            		result = getAllReaders(data);
            		break;
            	}
            	case 'a' :
            	{
            		try
                    {
                        result = addReader(data);
                    }
                    catch (Throwable t)
                    {
                        result = Boolean.FALSE;
                    }
            		break;
            	}
            	case 'e' :
            	{
            		try
            		{
            			result = removeReader(data);
            		}
            		catch (Throwable t)
            		{
            			result = Boolean.FALSE;    				
            		}
            		break;
            	}
            	case 'm' :
            	{
            		try
            		{
            			result = removeWriter(data);
            		}
            		catch (Throwable t)
            		{
            			result = Boolean.FALSE;    				
            		}
            		break;
            	}
            	case 'u' :
            	{
            		try
            		{
            			result = upgradeLock(data);
            		}
            		catch (Throwable t)
            		{
            			result = Boolean.FALSE;    				
            		}
            	}
            	case 's' :
            	{
            		try
            		{
            			result = setWriter(data);
            		}
            		catch (Throwable t)
            		{
            			result = Boolean.FALSE;    				
            		}
            		break;
            	}
            	case 'h' :
            	{
            		try
            		{
            			result = hasReadLock(data);
            		}
            		catch (Throwable t)
            		{
            			result = Boolean.FALSE;     				
            		}
            		break;
            	}
            	
            	default :
            	{
            		result = new PersistenceBrokerException("unknown command:" + selector); 		
            	}
            }
            
            ObjectOutputStream oos = new ObjectOutputStream(response.getOutputStream());
            oos.writeObject(result);
            oos.flush();
            oos.close();    
        }
        catch (Throwable t)
        {
            lastError = t;
            System.err.println(t.getMessage());
            t.printStackTrace(System.err);
        }
        
    }
    
    
    private LockEntry getWriter(byte[] data) throws IOException, ClassNotFoundException
    {
        LockEntry entry;
        Identity oid = (Identity) deserialize(data);    		
        entry = lockmap.getWriter(oid);
        return entry;
    }

	private Collection getAllReaders(byte[] data) throws IOException, ClassNotFoundException
	{
		Collection readers;
		Identity oid = (Identity) deserialize(data);   		
		readers = lockmap.getReaders(oid);
		
		return new ArrayList(readers);
	}

	private Boolean addReader(byte[] data) throws IOException, ClassNotFoundException
	{
		Boolean result = Boolean.TRUE;
		LockEntry lock = (LockEntry) deserialize(data); 		
        try
        {
            lockmap.addReaderInternal(lock);
        }
        catch (Throwable t)
        {
			result = Boolean.FALSE;
        }
		return result;
	}

	private Boolean removeReader(byte[] data) throws IOException, ClassNotFoundException
	{
		Boolean result = Boolean.TRUE;
		LockEntry lock = (LockEntry) deserialize(data); 		
		try
		{
			lockmap.removeReaderByLock(lock);
		}
		catch (Throwable t)
		{
			result = Boolean.FALSE;
		}
		return result;
	}

	private Boolean removeWriter(byte[] data) throws IOException, ClassNotFoundException
	{
		Boolean result = Boolean.TRUE;
		LockEntry lock = (LockEntry) deserialize(data); 		
		try
		{
			lockmap.removeWriter(lock);
		}
		catch (Throwable t)
		{
			result = Boolean.FALSE;
		}
		return result;
	}

	private Boolean upgradeLock(byte[] data) throws IOException, ClassNotFoundException
	{
		Boolean result = Boolean.TRUE;
		LockEntry lock = (LockEntry) deserialize(data); 		
		try
		{
			lockmap.upgradeLock(lock);
		}
		catch (Throwable t)
		{
			result = Boolean.FALSE;
		}
		return result;
	}

	private Boolean setWriter(byte[] data) throws IOException, ClassNotFoundException
	{
		Boolean result = Boolean.TRUE;
		LockEntry lock = (LockEntry) deserialize(data); 		
		try
		{
			lockmap.setWriterByLock(lock);
		}
		catch (Throwable t)
		{
			result = Boolean.FALSE;
		}
		return result;
	}

	private Boolean hasReadLock(byte[] data) throws IOException, ClassNotFoundException
	{
		Boolean result = Boolean.TRUE;
		LockEntry lock = (LockEntry) deserialize(data); 		
		try
		{
			lockmap.hasReadLock(lock);
		}
		catch (Throwable t)
		{
			result = Boolean.FALSE;
		}
		return result;
	}

	
	private Object deserialize(byte[] anArray) throws IOException, ClassNotFoundException
	{
		ByteArrayInputStream bais = new ByteArrayInputStream(anArray);
		ObjectInputStream ois = new ObjectInputStream(bais);
		Object result = ois.readObject();
		ois.close();
		bais.close();
		return result;
	}

    
    private byte[] getRequestByteArray(HttpServletRequest request)
        throws IOException
    {
        InputStream is = request.getInputStream();
        byte[] block = new byte[4096];
        byte[] req = null;
        int len = 0;
        
        len = is.read(block,0,block.length);
        if (len <= 0)
        {
        	return null;
        }
        else
        {
        	req = new byte[len];
        	System.arraycopy(block,0,req,0,len);
        }
        
        while (len > 0)
        {
			len = is.read(block,0,block.length);
			if (len <= 0)
			{
				break;
			}
			else
			{
				byte[] fresh = new byte[req.length + len];
				System.arraycopy(req,0,fresh,0,req.length);
				System.arraycopy(block,0,fresh,req.length,len);
				req = fresh;
			}
        }
        return req;
    }

}
