package org.apache.ojb.odmg.locking;

/* Copyright 2002-2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import org.apache.ojb.broker.Identity;
import org.apache.ojb.broker.PersistenceBroker;
import org.apache.ojb.broker.PersistenceBrokerException;
import org.apache.ojb.broker.query.Criteria;
import org.apache.ojb.broker.query.Query;
import org.apache.ojb.broker.query.QueryFactory;
import org.apache.ojb.broker.util.configuration.Configuration;
import org.apache.ojb.broker.util.configuration.ConfigurationException;
import org.apache.ojb.broker.util.logging.Logger;
import org.apache.ojb.broker.util.logging.LoggerFactory;
import org.apache.ojb.odmg.TransactionImpl;
import org.apache.ojb.odmg.TxManagerFactory;

/**
 * @deprecated Using this class is not safe! Please use the RemoteLockMapImpl instead.
 */
public class PersistentLockMapImpl implements LockMap
{
    private Logger log = LoggerFactory.getLogger(PersistentLockMapImpl.class);

    /**
     * obtain a PersistenceBroker instance for persistence operations.
     */
    private PersistenceBroker getBroker()
    {
        return TxManagerFactory.instance().getCurrentTransaction().getBroker();
    }

    /**
     * remove a lock entry from the OJB cache to enforce
     * synchronization with the database.
     */
    private void unCache(LockEntry entry)
    {
        if (entry != null)
        {
            try
            {
                getBroker().removeFromCache(entry);
            }
            catch (PersistenceBrokerException e)
            {
                log.error("Could not remove entry from cache " + entry, e);
            }
        }
    }


    /**
     * remove a collection of lock entries from the OJB cache to enforce
     * synchronization with the database.
     */
    private void unCache(Collection col)
    {
        Iterator i = col.iterator();
        while (i.hasNext())
        {
            try
            {
                getBroker().removeFromCache(i.next());
            }
            catch (PersistenceBrokerException e)
            {
                log.error("Could not remove entry from cache", e);
            }
        }
    }


    /**
     * returns the LockEntry for the Writer of object obj.
     * If now writer exists, null is returned.
     */
    public LockEntry getWriter(Object obj)
    {
        removeTimedOutLocks(AbstractLockStrategy.DEFAULT_LOCK_TIMEOUT);
        /* @todo smarter solution in future */
        // fix/workaround
        // When Identity needs new id's we must overgive
        // the a target broker when run with multiple databases
        // using H/L sequence manager
        PersistenceBroker broker = getBroker();
        Identity oid = new Identity(obj, broker);
        Criteria criteria = new Criteria();
        criteria.addEqualTo("oidString", oid.toString());
        criteria.addEqualTo("lockType", new Integer(LockEntry.LOCK_WRITE));
        Query q = QueryFactory.newQuery(LockEntry.class, criteria);
        LockEntry result = null;
        try
        {
            result = (LockEntry) getBroker().getObjectByQuery(q);
        }
        catch (PersistenceBrokerException e)
        {
            log.error(e);
        }
        unCache(result);
        return result;
    }

    /**
     * returns a collection of Reader LockEntries for object obj.
     * If now LockEntries could be found an empty Vector is returned.
     */
    public Collection getReaders(Object obj)
    {
        removeTimedOutLocks(AbstractLockStrategy.DEFAULT_LOCK_TIMEOUT);
        Identity oid = new Identity(obj, getBroker());
        Criteria criteria = new Criteria();
        criteria.addEqualTo("oidString", oid.toString());
        criteria.addEqualTo("lockType", new Integer(LockEntry.LOCK_READ));
        Query q = QueryFactory.newQuery(LockEntry.class, criteria);
        try
        {
            Collection result = getBroker().getCollectionByQuery(q);
            unCache(result);
            return result;
        }
        catch (PersistenceBrokerException e)
        {
            log.error("Cannot obtain collection of LockEntries, query was " + q, e);
            return new Vector();
        }
    }


    /**
     * Add a reader lock entry for transaction tx on object obj
     * to the persistent storage.
     */
    public boolean addReader(TransactionImpl tx, Object obj)
    {
        removeTimedOutLocks(AbstractLockStrategy.DEFAULT_LOCK_TIMEOUT);
        try
        {
            LockEntry lock = new LockEntry(new Identity(obj,getBroker()).toString(),
                    tx.getGUID(),
                    System.currentTimeMillis(),
                    LockStrategyFactory.getIsolationLevel(obj),
                    LockEntry.LOCK_READ);

            getBroker().store(lock);
            unCache(lock);
            return true;
        }
        catch (PersistenceBrokerException e)
        {
            log.error("Cannot store LockEntry for object " + obj + " in transaction " + tx);
            return false;
        }
    }


    /**
     * remove a reader lock entry for transaction tx on object obj
     * from the persistent storage.
     */
    public void removeReader(TransactionImpl tx, Object obj)
    {
        try
        {
            LockEntry lock = new LockEntry(new Identity(obj,getBroker()).toString(), tx.getGUID());
            getBroker().delete(lock);
            unCache(lock);
        }
        catch (PersistenceBrokerException e)
        {
            log.error("Cannot remove LockEntry for object " + obj + " in transaction " + tx);
        }
    }

    /**
     * remove a writer lock entry for transaction tx on object obj
     * from the persistent storage.
     */
    public void removeWriter(LockEntry writer)
    {
        try
        {
            getBroker().delete(writer);
            unCache(writer);
        }
        catch (PersistenceBrokerException e)
        {
            log.error("Cannot remove LockEntry", e);
        }
    }

    /**
     * upgrade a reader lock entry for transaction tx on object obj
     * and write it to the persistent storage.
     */
    public boolean upgradeLock(LockEntry reader)
    {
        reader.setLockType(LockEntry.LOCK_WRITE);

        try
        {
            getBroker().store(reader);
            unCache(reader);
            return true;
        }
        catch (PersistenceBrokerException e)
        {
            log.error("Cannot upgrade LockEntry " + reader, e);
            return false;
        }
    }

    /**
     * generate a writer lock entry for transaction tx on object obj
     * and write it to the persistent storage.
     */
    public boolean setWriter(TransactionImpl tx, Object obj)
    {
        removeTimedOutLocks(AbstractLockStrategy.DEFAULT_LOCK_TIMEOUT);
        try
        {
            LockEntry lock = new LockEntry(new Identity(obj,getBroker()).toString(),
                    tx.getGUID(),
                    System.currentTimeMillis(),
                    LockStrategyFactory.getIsolationLevel(obj),
                    LockEntry.LOCK_WRITE);

            getBroker().store(lock);
            unCache(lock);
            return true;
        }
        catch (PersistenceBrokerException e)
        {
            log.error("Cannot set LockEntry for object " + obj + " in transaction " + tx);
            return false;
        }
    }

    /**
     * check if there is a reader lock entry for transaction tx on object obj
     * in the persistent storage.
     */
    public boolean hasReadLock(TransactionImpl tx, Object obj)
    {
        removeTimedOutLocks(AbstractLockStrategy.DEFAULT_LOCK_TIMEOUT);
        try
        {
            LockEntry lock = new LockEntry(new Identity(obj,getBroker()).toString(), tx.getGUID());
            Identity oid = new Identity(lock,getBroker());
            LockEntry result = (LockEntry) getBroker().getObjectByIdentity(oid);
            unCache(result);

            return (result != null);
        }
        catch (PersistenceBrokerException e)
        {
            log.error("Cannot check read lock for object " + obj + " in transaction " + tx, e);
            return false;
        }
    }


    /**
     * removes all timed out lock entries from the persistent storage.
     * The timeout value can be set in the OJB properties file.
     */
    protected void removeTimedOutLocks(long timeout)
    {
        int count = 0;
        long maxAge = System.currentTimeMillis() - timeout;
        Criteria criteria = new Criteria();
        criteria.addLessThan("timestamp", new Long(maxAge));
        Query q = QueryFactory.newQuery(LockEntry.class, criteria);
        Iterator i = getBroker().getIteratorByQuery(q);
        while (i.hasNext())
        {
            try
            {
                getBroker().delete(i.next());
                count++;
            }
            catch (PersistenceBrokerException e)
            {
                log.error("Removing timed out lock failed", e);
            }
        }
        if (count > 0)
        {
            if (log.isDebugEnabled()) log.debug("Removed " + count + " timedout locks");
        }
    }
    /* (non-Javadoc)
     * @see org.apache.ojb.broker.util.configuration.Configurable#configure(org.apache.ojb.broker.util.configuration.Configuration)
     */
    public void configure(Configuration pConfig) throws ConfigurationException
    {
        // noop

    }

}
