package org.apache.ojb.broker;

import org.apache.ojb.broker.core.PersistenceBrokerConfiguration;
import org.apache.ojb.broker.metadata.ClassDescriptor;
import org.apache.ojb.broker.metadata.ObjectReferenceDescriptor;
import org.apache.ojb.broker.query.QueryByIdentity;
import org.apache.ojb.broker.query.QueryFactory;
import org.apache.ojb.broker.util.configuration.ConfigurationException;
import org.apache.ojb.junit.PBTestCase;

/**
 * Demo Application that shows basic concepts for Applications using the PersistenceBroker
 * as a mediator for persistence
 */
public class BrokerExamples extends PBTestCase
{
    public static void main(String[] args)
    {
        String[] arr = {BrokerExamples.class.getName()};
        junit.textui.TestRunner.main(arr);
    }

    private int testId = 88881;

    public BrokerExamples(String name)
    {
        super(name);
    }

    protected Article createArticle(int id)
    {
        Article a = new Article();
        a.setArticleId(id);
        a.setArticleName("New Funny Article " + id);
        a.setIsSelloutArticle(true);
        a.setMinimumStock(100);
        a.setOrderedUnits(17);
        a.setPrice(0.45);
        a.setProductGroupId(1);
        a.setStock(234);
        a.setSupplierId(4);
        a.setUnit("bottle");
        ProductGroup tmpPG = new ProductGroup();
        tmpPG.setId(1);
        Identity pgID = new Identity(tmpPG, broker);
        ProductGroupProxy pgProxy = new ProductGroupProxy(broker.getPBKey(), pgID);
        a.setProductGroup(pgProxy);
        return a;
    }

    public void testCollectionRetrieval() throws Exception
    {
        // Use PersistenceBroker to handle persistence of Business objects

        // loop through categories 1 to 9 and print them
        // A ProductGroup holds a Vector of all Article Objects in the specific category
        // the repository.xml specifies that the Vector of CArtikels has to be materialized immediately.
        // Thus printing the category will also print all Articles in the category
        for (int i = 1; i < 9; i++)
        {
            ProductGroup example = new ProductGroup();
            example.setId(i);

            ProductGroup group =
                    (ProductGroup) broker.getObjectByQuery(QueryFactory.newQuery(example));
            assertEquals("should be equal", i, group.getId());
        }
    }

    public void testModifications() throws Exception
    {
        ProductGroup pg;
        pg = new ProductGroup();
        pg.setId(1);

        //create a new Article and play with it
        Article article = new Article();
        article.setPrice(2);
        article.setStock(0);
        article.setArticleId(testId);
        article.setProductGroup(pg);

        broker.beginTransaction();
        for (int i = 1; i < 50; i++)
        {
            article.addToStock(10);
            broker.store(article);
            broker.delete(article);
            broker.store(article);
        }
        broker.commitTransaction();

        Identity oid = new Identity(article, broker);
        Object result = broker.getObjectByIdentity(oid);
        assertNotNull(result);
    }

    /**
     * here we see what a good job the ObjectCache does:
     * if the Cache has not been flushed by the GC, All previously loaded Objects are still hot.
     * I.e. they are reloaded from the Cache and not materialized from the RDBMS
     */
    public void testObjectCache() throws Exception
    {
        Article tmpArticle = createArticle(testId);
        Identity tmpOID = new Identity(tmpArticle, broker);
        broker.beginTransaction();
        broker.store(tmpArticle);
        broker.commitTransaction();
        broker.clearCache();
        Article cached =
                (Article) broker.serviceObjectCache().lookup(tmpOID);
        assertNull("After flushing Object should be not in cache", cached);
        Article article = (Article) broker.getObjectByQuery(new QueryByIdentity(tmpOID));
        cached = (Article) broker.serviceObjectCache().lookup(tmpOID);
        assertNotNull("now object should be found in cache", cached);
        assertEquals("should be the same Identity", new Identity(cached, broker), new Identity(article,broker));
        assertEquals("objects should have identical values", cached.toString(), article.toString());
    }

    /**
     * Insert the method's description here.
     * Creation date: (06.12.2000 22:15:33)
     * @param id int
     */
    protected Article createSimpleArticle(int id)
    {
        Article a = new Article();
        a.setArticleId(id);
        a.setArticleName("New Funny Article " + id);
        a.setIsSelloutArticle(false);
        a.setMinimumStock(100);
        a.setOrderedUnits(17);
        a.setPrice(0.45);
        a.setProductGroupId(1);
        a.setStock(234);
        a.setSupplierId(4);
        a.setUnit("bottle");
        return a;
    }

    /**
     * here we see what a good job the ObjectCache does:
     * if the Cache has not been flushed by the GC, All previously loaded Objects are still hot.
     * I.e. they are reloaded from the Cache and not materialized from the RDBMS
     */
    public void testShallowAndDeepRetrieval() throws Exception
    {
        // this test only makes sense in singlevm mode !
        PersistenceBrokerConfiguration config = null;
        try
        {
            config =
                    (PersistenceBrokerConfiguration) PersistenceBrokerFactory
                    .getConfigurator()
                    .getConfigurationFor(
                            null);
        }
        catch (ConfigurationException e)
        {
            fail(e.getMessage());
        }

            try
            {
                // ensure there is an item to find
                Article tmpArticle = createArticle(testId);
                Identity tmpOID = new Identity(tmpArticle, broker);
                broker.beginTransaction();
                broker.store(tmpArticle);
                broker.commitTransaction();
                broker.clearCache();

                // switch to shallow retrieval
                ClassDescriptor cld = broker.getClassDescriptor(Article.class);
                ObjectReferenceDescriptor ord =
                        (ObjectReferenceDescriptor) cld.getObjectReferenceDescriptors().get(0);
                ord.setCascadeRetrieve(false);
                // should work without setting cld
                // broker.setClassDescriptor(cld);

                Article article = (Article) broker.getObjectByIdentity(tmpOID);
                assertNull("now reference should be null", article.getProductGroup());

                // now switch to deep retrieval
                ord.setCascadeRetrieve(true);
                // should work without setting cld
                // broker.setClassDescriptor(cld);
                broker.clearCache();
                article = (Article) broker.getObjectByIdentity(tmpOID);
                assertNotNull("now reference should NOT be null", article.getProductGroup());

            }
            catch (Exception e)
            {
                e.printStackTrace();
                throw e;
            }

    }


    /**
     * tests the PB.retrieveReference() feature
     */
    public void testRetrieveReference() throws Exception
    {
        // this test only makes sense in singlevm mode !
        PersistenceBrokerConfiguration config = null;
        try
        {
            config =
                    (PersistenceBrokerConfiguration) PersistenceBrokerFactory
                    .getConfigurator()
                    .getConfigurationFor(
                            null);
        }
        catch (ConfigurationException e)
        {
            fail(e.getMessage());
        }

        // ensure there is an item to find
        Article tmpArticle = createArticle(testId);
        Identity tmpOID = new Identity(tmpArticle, broker);
        broker.beginTransaction();
        broker.store(tmpArticle);
        broker.commitTransaction();
        broker.clearCache();

        // switch to shallow retrieval
        ClassDescriptor cld = broker.getClassDescriptor(Article.class);
        // article only has one ord
        ObjectReferenceDescriptor ord =
                (ObjectReferenceDescriptor) cld.getObjectReferenceDescriptors().get(0);
        ord.setCascadeRetrieve(false);
        // should work without setting cld
        // broker.setClassDescriptor(cld);

        Article article = (Article) broker.getObjectByIdentity(tmpOID);
        assertNull("now reference should be null", article.getProductGroup());

        // now force loading:
        broker.retrieveReference(article, "productGroup");
        assertNotNull("now reference should NOT be null", article.getProductGroup());

        // repair cld
        ord.setCascadeRetrieve(true);
        // should work without setting cld
        // broker.setClassDescriptor(cld);

    }

    /**
     * tests the PB.retrieveAllReferences() feature
     */
    public void testRetrieveAllReferences()
    {
        // this test only makes sense in singlevm mode !
        PersistenceBrokerConfiguration config = null;
        try
        {
            config =
                    (PersistenceBrokerConfiguration) PersistenceBrokerFactory
                    .getConfigurator()
                    .getConfigurationFor(
                            null);
        }
        catch (ConfigurationException e)
        {
            fail(e.getMessage());
        }

        // ensure there is an item to find
        Article tmpArticle = createArticle(testId);
        Identity tmpOID = new Identity(tmpArticle, broker);
        broker.beginTransaction();
        broker.store(tmpArticle);
        broker.commitTransaction();
        broker.clearCache();

        // switch to shallow retrieval
        ClassDescriptor cld = broker.getClassDescriptor(Article.class);
        ObjectReferenceDescriptor ord =
                (ObjectReferenceDescriptor) cld.getObjectReferenceDescriptors().get(0);
        ord.setCascadeRetrieve(false);
        // should work without setting cld
        // broker.setClassDescriptor(cld);

        Article article = (Article) broker.getObjectByIdentity(tmpOID);
        assertNull("now reference should be null", article.getProductGroup());

        // now force loading:
        broker.retrieveAllReferences(article);
        assertNotNull("now reference should NOT be null", article.getProductGroup());

        // clean up cld
        ord.setCascadeRetrieve(true);
        // should work without setting cld
        // broker.setClassDescriptor(cld);

    }
}
