package org.apache.ojb.broker;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.ojb.broker.metadata.ClassDescriptor;
import org.apache.ojb.broker.metadata.CollectionDescriptor;
import org.apache.ojb.broker.metadata.ObjectReferenceDescriptor;
import org.apache.ojb.broker.query.Criteria;
import org.apache.ojb.broker.query.Query;
import org.apache.ojb.broker.query.QueryFactory;
import org.apache.ojb.junit.PBTestCase;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * Test (non-decomposed) M:N relations.
 *
 * IMPORTANT NOTE: The global runtime metadata changes made by this test case
 * are NOT recommended in multithreaded environments, because they are global
 * and each thread will be affected.
 *
 * @author <a href="mailto:arminw@apache.org">Armin Waibel</a>
 * @version $Id: M2NTest.java,v 1.7 2004/06/22 18:16:22 arminw Exp $
 */
public class M2NTest extends PBTestCase
{
    static final int NONE = ObjectReferenceDescriptor.CASCADE_NONE;
    static final int LINK = ObjectReferenceDescriptor.CASCADE_LINK;
    static final int OBJECT = ObjectReferenceDescriptor.CASCADE_OBJECT;

    public static void main(String[] args)
    {
        junit.textui.TestRunner.main(new String[]{M2NTest.class.getName()});
    }

    public void tearDown() throws Exception
    {
        if(broker != null)
        {
            changeActorCollectionDescriptorTo(false, NONE, NONE, false);
            changeMovieCollectionDescriptorTo(false, NONE, NONE, false);
            broker.close();
        }
        super.tearDown();
    }


    public void testAutoUpdateDeleteSettings()
    {
        changeActorCollectionDescriptorTo(false, false, false, false);
        CollectionDescriptor ord = broker.getClassDescriptor(Actor.class)
                .getCollectionDescriptorByName("movies");
        assertEquals(CollectionDescriptor.CASCADE_LINK, ord.getCascadingStore());
        assertEquals(CollectionDescriptor.CASCADE_LINK, ord.getCascadingDelete());
        assertEquals(false, ord.getCascadeStore());
        assertEquals(false, ord.getCascadeDelete());

        changeActorCollectionDescriptorTo(false, true, true, false);
        ord = broker.getClassDescriptor(Actor.class)
                .getCollectionDescriptorByName("movies");
        assertEquals(CollectionDescriptor.CASCADE_OBJECT, ord.getCascadingStore());
        assertEquals(CollectionDescriptor.CASCADE_OBJECT, ord.getCascadingDelete());
        assertEquals(true, ord.getCascadeStore());
        assertEquals(true, ord.getCascadeDelete());
    }

    //============================================================
    // auto-update / auto-retrieve false
    //============================================================

    /**
     * All auto-xxx settings are false, thus expect that all thing must be
     * done by hand
     */
    public void testStoreFFFF()
    {
        changeMovieCollectionDescriptorTo(false, NONE, NONE, false);
        changeActorCollectionDescriptorTo(false, NONE, NONE, false);
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActors(postfix);
        doTestStoreFFFX(movie, postfix);
    }

    /**
     * All auto-xxx settings are false, thus expect that all thing must be
     * done by hand, proxy true should not affect anything
     */
    public void testStoreFFFT()
    {
        changeMovieCollectionDescriptorTo(false, NONE, NONE, true);
        changeActorCollectionDescriptorTo(false, NONE, NONE, false);
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActors(postfix);
        doTestStoreFFFX(movie, postfix);
    }

    /**
     * All auto-xxx settings are false, thus expect that all thing must be
     * done by hand. Actors has a collection of movies too - should be ignored
     * in this case.
     */
    public void testStoreFFFF_2()
    {
        changeMovieCollectionDescriptorTo(false, NONE, NONE, false);
        changeActorCollectionDescriptorTo(false, NONE, NONE, false);
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActorsAndBackReferences(postfix);
        doTestStoreFFFX(movie, postfix);
    }

    /**
     * All auto-xxx settings are false, thus expect that all thing must be
     * done by hand, proxy true should not affect anything. Actors has a
     * collection of movies too - should be ignored in this case.
     */
    public void testStoreFFFT_2()
    {
        changeMovieCollectionDescriptorTo(false, NONE, NONE, true);
        changeActorCollectionDescriptorTo(false, NONE, NONE, false);
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActorsAndBackReferences(postfix);
        doTestStoreFFFX(movie, postfix);
    }

    public void doTestStoreFFFX(Movie movie, String postfix)
    {
        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();

        /*
        all auto-xxx settings are false, so only the movie object should be
        stored - no other inserts!
        */
        Query queryMovie = movieQuery(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        Query queryActor = actorQuery(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        // auto-update is false, thus we don't expect an Actor object
        assertEquals(0, resultActor.size());

        Query queryRole = roleQuery(null, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        // auto-update is false, thus we don't expect Role objects
        assertEquals(0, resultRole.size());

        broker.beginTransaction();
        /*
        now we store the right-side objects and the intermediary entries
        */
        Iterator it = movie.getActors().iterator();
        while(it.hasNext())
        {
            Object actor = it.next();
            broker.store(actor);
        }
        // now both side exist and we can link the references
        broker.serviceBrokerHelper().link(movie, "actors", true);
        /*
        alternative call
        broker.serviceBrokerHelper().link(movie, true);
        */
        broker.commitTransaction();

        /*
        now we expect all stored objects
        */
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        Movie readMovie = (Movie) broker.getObjectByIdentity(oid);
        assertNotNull(readMovie);
        // auto-retrieve false
        assertTrue(readMovie.getActors() == null || readMovie.getActors().size() == 0);

        broker.retrieveAllReferences(readMovie);
        assertEquals(3, readMovie.getActors().size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());
        Actor readActor = (Actor) resultActor.iterator().next();
        // auto-retrieve false
        assertTrue(readActor.getMovies() == null || readActor.getMovies().size() == 0);
        broker.retrieveAllReferences(readActor);
        assertEquals(1, readActor.getMovies().size());

        // We try to delete all objects
        // first do unlink the m:n references
        broker.beginTransaction();
        broker.serviceBrokerHelper().unlink(readMovie, "actors");
        /*
        alternative call
        broker.serviceBrokerHelper().unlink(readMovie);
        */
        broker.commitTransaction();

        broker.clearCache();
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());

        // now we delete the n- and m-side objects
        broker.beginTransaction();
        Iterator iter = movie.getActors().iterator();
        while(iter.hasNext())
        {
            broker.delete(iter.next());
        }
        broker.delete(movie);
        broker.commitTransaction();

        broker.clearCache();
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(0, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(0, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());
    }




    //============================================================
    // auto-retrieve true / auto-update/auto-delete false
    //============================================================
    /**
     * auto-retrieve is true, other false. proxy true/false should not affect
     * anything.
     */
    public void testStoreTFFF()
    {
        changeMovieCollectionDescriptorTo(true, NONE, NONE, false);
        changeActorCollectionDescriptorTo(true, NONE, NONE, false);
        doTestStoreTFFX();
    }

    /**
     * auto-retrieve is true, other false. proxy true/false should not affect
     * anything.
     */
    public void testStoreTFFT()
    {
        changeMovieCollectionDescriptorTo(true, NONE, NONE, true);
        changeActorCollectionDescriptorTo(true, NONE, NONE, false);
        doTestStoreTFFX();
    }

    public void doTestStoreTFFX()
    {
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActorsAndBackReferences(postfix);

        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();

        /*
        auto-update settings is false, so only the movie object should be
        stored - no other inserts!
        */
        Query queryMovie = movieQuery(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        Query queryActor = actorQuery(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        // auto-update is false, thus we don't expect an Actor object
        assertEquals(0, resultActor.size());

        Query queryRole = roleQuery(null, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        // auto-update is false, thus we don't expect Role objects
        assertEquals(0, resultRole.size());

        broker.beginTransaction();
        /*
        now we store the right-side objects and the intermediary entries
        */
        Iterator it = movie.getActors().iterator();
        while(it.hasNext())
        {
            Object actor = it.next();
            broker.store(actor);
        }
        // now both side exist and we can link the references
        broker.serviceBrokerHelper().link(movie, "actors", true);
        /*
        alternative call
        broker.serviceBrokerHelper().link(movie, true);
        */
        broker.commitTransaction();

        /*
        now we expect all stored objects
        */
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        Movie readMovie = (Movie) broker.getObjectByIdentity(oid);
        assertNotNull(readMovie);
        // auto-retrieve true
        assertTrue(readMovie.getActors() != null);
        assertEquals(3, readMovie.getActors().size());

        // We try to delete all objects
        // first do unlink the m:n references
        broker.beginTransaction();
        broker.serviceBrokerHelper().unlink(readMovie, "actors");
        /*
        alternative call
        broker.serviceBrokerHelper().unlink(readMovie);
        */
        broker.commitTransaction();

        broker.clearCache();
        // TODO: replace this with query below (when prefetching bug was solved)
        Movie movieLookup = (Movie) broker.getObjectByIdentity(new Identity(movie, broker));
        assertNotNull(movieLookup);
        //resultMovie = broker.getCollectionByQuery(queryMovie);
        //assertEquals(1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());

        // now we delete the n- and m-side objects
        broker.beginTransaction();
        Iterator iter = movie.getActors().iterator();
        while(iter.hasNext())
        {
            broker.delete(iter.next());
        }
        broker.delete(movie);
        broker.commitTransaction();

        broker.clearCache();
        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(0, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(0, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());
    }




    //============================================================
    // auto-update / auto-retrieve true
    //============================================================

    public void testStoreTTFF()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, NONE, false);
        changeActorCollectionDescriptorTo(true, OBJECT, NONE, false);
        doTestStoreTTXX();
    }

    public void testStoreTTFT()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, NONE, true);
        changeActorCollectionDescriptorTo(true, OBJECT, NONE, false);
        doTestStoreTTXX();
    }

    public void doTestStoreTTXX()
    {
        String postfix = "" + System.currentTimeMillis();
        Movie movie = buildMovieWithActors(postfix);

        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();

        Query queryMovie = movieQuery(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        Query queryActor = actorQuery(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        Query queryRole = roleQuery(null, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        Movie readMovie = (Movie) broker.getObjectByIdentity(oid);
        assertNotNull(readMovie);
        assertEquals(3, readMovie.getActors().size());
    }

    public void testStoreUpdateTTFF()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, NONE, false);
        changeActorCollectionDescriptorTo(true, OBJECT, NONE, false);
        doTestStoreUpdateTTXX();
    }

    public void testStoreUpdateTTFF_2()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, NONE, true);
        changeActorCollectionDescriptorTo(true, OBJECT, NONE, false);
        doTestStoreUpdateTTXX();
    }

    public void doTestStoreUpdateTTXX()
    {
        String postfix = "doTestStoreUpdateTTXX_" + System.currentTimeMillis();
        Movie movie = buildMovieWithActors(postfix);

        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();

        Query queryMovie = movieQuery(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        Query queryActor = actorQuery(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        Query queryRole = roleQuery(null, movie);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        //*****************************
        // remove all actors
        movie.setActors(new ArrayList());
        broker.beginTransaction();
        broker.store(movie);
        broker.commitTransaction();
        //*****************************

        broker.clearCache();
        Identity oid = new Identity(movie, broker);
        movie = (Movie) broker.getObjectByIdentity(oid);

        resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(1, resultMovie.size());

        resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(3, resultActor.size());

        resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(0, resultRole.size());

        assertNotNull(movie);
        assertEquals(0, movie.getActors().size());
    }

    public void testStoreUpdateActorTTFF()
    {
        changeMovieCollectionDescriptorTo(true, OBJECT, NONE, false);
        changeActorCollectionDescriptorTo(true, OBJECT, NONE, false);

        String postfix = "" + System.currentTimeMillis();
        Actor actor = buildActorWithMovies(postfix);

        broker.beginTransaction();
        broker.store(actor);
        broker.commitTransaction();

        broker.clearCache();
        Query queryMovie = movieQuery(postfix);
        Collection resultMovie = broker.getCollectionByQuery(queryMovie);
        assertEquals(3, resultMovie.size());

        Query queryActor = actorQuery(postfix);
        Collection resultActor = broker.getCollectionByQuery(queryActor);
        assertEquals(1, resultActor.size());

        Query queryRole = roleQuery(actor, null);
        Collection resultRole = broker.getCollectionByQuery(queryRole);
        assertEquals(3, resultRole.size());

        broker.clearCache();
        Identity oid = new Identity(actor, broker);
        Actor loadedActor = (Actor) broker.getObjectByIdentity(oid);
        assertNotNull(loadedActor);
        MovieManageableCollection col = loadedActor.getMovies();
        assertNotNull(col);
        Movie movie = col.get(0);
    }

    //=======================================================================
    // helper methods
    //=======================================================================
    Query movieQuery(String postfix)
    {
        Criteria c = new Criteria();
        c.addLike("idStr", "%" + postfix + "%");
        Query q = QueryFactory.newQuery(Movie.class, c);
        return q;
    }

    Query actorQuery(String postfix)
    {
        Criteria c = new Criteria();
        c.addLike("name", "%" + postfix + "%");
        Query q = QueryFactory.newQuery(Actor.class, c);
        return q;
    }

    Query roleQuery(Actor actor, Movie movie)
    {
        Criteria c = new Criteria();
        if(actor != null) c.addEqualTo("actorId", actor.getId());
        if(movie != null)
        {
            c.addEqualTo("movieIntId", movie.getIdInt());
            c.addEqualTo("movieStrId", movie.getIdStr());
        }
        Query q = QueryFactory.newQuery(Role.class, c);
        return q;
    }

    /**
     * Returns 1 movie object with 3 actor objects
     */
    Movie buildMovieWithActors(String postfixId)
    {
        Movie m = new MovieImpl(postfixId,
                "Dr. Strangelove or: How I Learned to Stop Worrying and Love the Bomb " + postfixId,
                "An insane general starts a process to nuclear holocaust that a war" +
                " room of politicians and generals frantically try to stop. " + postfixId);
        Actor a1 = new Actor("Peter Sellers " + postfixId);
        Actor a2 = new Actor("George C. Scott " + postfixId);
        Actor a3 = new Actor("Sterling Hayden " + postfixId);
        ArrayList list = new ArrayList();
        list.add(a1);
        list.add(a2);
        list.add(a3);
        m.setActors(list);
        return m;
    }

    /**
     * Returns 1 movie object with 3 actor objects and one actor object with
     * back-reference to movie object + 2 new movies
     */
    Movie buildMovieWithActorsAndBackReferences(String postfixId)
    {
        Movie m = new MovieImpl(postfixId,
                "Dr. Strangelove or: How I Learned to Stop Worrying and Love the Bomb " + postfixId,
                "An insane general starts a process to nuclear holocaust that a war" +
                " room of politicians and generals frantically try to stop. " + postfixId);
        Actor a1 = new Actor("Peter Sellers " + postfixId);
        Actor a2 = new Actor("George C. Scott " + postfixId);
        Actor a3 = new Actor("Sterling Hayden " + postfixId);
        ArrayList list = new ArrayList();
        list.add(a1);
        list.add(a2);
        list.add(a3);
        m.setActors(list);

        Movie m1 = new MovieImpl(postfixId + "", "A Shot in the Dark",
                "As murder follows murder, beautiful Maria is the obvious suspect...");
        Movie m2 = new MovieImpl(postfixId + "", "The Pink Panther",
                " In the first movie starring Peter Sellers as the bumbling Inspector Clouseau...");

        MovieManageableCollection mlist1 = new MovieManageableCollection();
        mlist1.add(m);
        mlist1.add(m1);
        mlist1.add(m2);
        MovieManageableCollection mlist2 = new MovieManageableCollection();
        MovieManageableCollection mlist3 = new MovieManageableCollection();
        a1.setMovies(mlist1);
        a2.setMovies(mlist2);
        a3.setMovies(mlist3);

        return m;
    }

    Actor buildActorWithMovies(String postfixId)
    {
        Actor a = new Actor("John Cusack" + postfixId);
        MovieManageableCollection list = new MovieManageableCollection();
        list.add(new MovieImpl("a_" + postfixId, "High Fidelity", "A comedy about fear of commitment, hating your job..." + postfixId));
        list.add(new MovieImpl("b_" + postfixId, "Identity", "When a nasty storm hits a hotel, ten strangers are stranded within ..." + postfixId));
        list.add(new MovieImpl("c_" + postfixId, "Grosse Pointe Blank", "Martin Blank is a professional assassin. He is sent on a mission to a small Detroit ..." + postfixId));
        a.setMovies(list);
        return a;
    }

    void changeActorCollectionDescriptorTo(boolean autoRetrieve, int autoUpdate, int autoDelete, boolean proxy)
    {
        ClassDescriptor cld = broker.getClassDescriptor(Actor.class);
        CollectionDescriptor cod = (CollectionDescriptor) cld.getCollectionDescriptors().get(0);
        cod.setLazy(proxy);
        cod.setCascadeRetrieve(autoRetrieve);
        cod.setCascadingStore(autoUpdate);
        cod.setCascadingDelete(autoDelete);
    }

    void changeActorCollectionDescriptorTo(boolean autoRetrieve, boolean autoUpdate, boolean autoDelete, boolean proxy)
    {
        ClassDescriptor cld = broker.getClassDescriptor(Actor.class);
        CollectionDescriptor cod = (CollectionDescriptor) cld.getCollectionDescriptors().get(0);
        cod.setLazy(proxy);
        cod.setCascadeRetrieve(autoRetrieve);
        cod.setCascadeStore(autoUpdate);
        cod.setCascadeDelete(autoDelete);
    }

    void changeMovieCollectionDescriptorTo(boolean autoRetrieve, int autoUpdate, int autoDelete, boolean proxy)
    {
        ClassDescriptor cld = broker.getClassDescriptor(MovieImpl.class);
        CollectionDescriptor cod = (CollectionDescriptor) cld.getCollectionDescriptors().get(0);
        cod.setLazy(proxy);
        cod.setCascadeRetrieve(autoRetrieve);
        cod.setCascadingStore(autoUpdate);
        cod.setCascadingDelete(autoDelete);
    }


    //=======================================================================
    // Inner classes, persistence capable  test classes
    //=======================================================================
    public static class MovieManageableCollection implements ManageableCollection
    {
        private ArrayList list = new ArrayList();

        public void ojbAdd(Object anObject)
        {
            list.add(anObject);
        }

        public void ojbAddAll(ManageableCollection otherCollection)
        {
            Iterator it = otherCollection.ojbIterator();
            while(it.hasNext())
            {
                list.add(it.next());
            }
        }

        public Iterator ojbIterator()
        {
            return list.iterator();
        }

        public void afterStore(PersistenceBroker broker) throws PersistenceBrokerException
        {
        }

        public Iterator iterator()
        {
            return list.iterator();
        }

        public int size()
        {
            return list.size();
        }

        public boolean isEmpty()
        {
            return list.isEmpty();
        }

        public void clear()
        {
            list.clear();
        }

        public boolean add(Movie movie)
        {
            return list.add(movie);
        }

        public boolean remove(Movie movie)
        {
            return list.remove(movie);
        }

        public boolean contains(Movie movie)
        {
            return list.contains(movie);
        }

        public Movie get(int index)
        {
            return (Movie) list.get(index);
        }
    }

    public static class Actor
    {
        private Integer id;
        private Integer id2;
        private String name;
        private MovieManageableCollection movies;

        public Actor()
        {
        }

        public Actor(String name)
        {
            this.name = name;
        }

        public MovieManageableCollection getMovies()
        {
            return movies;
        }

        public void setMovies(MovieManageableCollection movies)
        {
            this.movies = movies;
        }

        public Integer getId()
        {
            return id;
        }

        public void setId(Integer id)
        {
            this.id = id;
        }

        public Integer getId2()
        {
            return id2;
        }

        public void setId2(Integer id2)
        {
            this.id2 = id2;
        }

        public String getName()
        {
            return name;
        }

        public void setName(String name)
        {
            this.name = name;
        }

        public String toString()
        {
            return ToStringBuilder.reflectionToString(this).toString();
        }
    }

    public static interface Movie
    {
        public Collection getActors();

        public void setActors(Collection actors);

        public Integer getIdInt2();
        public Integer getIdInt();

        public void setIdInt2(Integer id2Int);
        public void setIdInt(Integer idInt);

        public String getIdStr();

        public void setIdStr(String idStr);

        public String getTitle();

        public void setTitle(String title);

        public String getDescription();

        public void setDescription(String description);
    }

    public static class MovieImpl implements Movie
    {
        private Integer idInt;
        private Integer idInt2;
        private String idStr;
        private String title;
        private String description;
        private Collection actors;

        public MovieImpl()
        {
        }

        public MovieImpl(String idStr, String title, String description)
        {
            this.idStr = idStr;
            this.title = title;
            this.description = description;
        }

        public Collection getActors()
        {
            return actors;
        }

        public void setActors(Collection actors)
        {
            this.actors = actors;
        }

        public Integer getIdInt()
        {
            return idInt;
        }

        public void setIdInt(Integer idInt)
        {
            this.idInt = idInt;
        }

        public Integer getIdInt2()
        {
            return idInt2;
        }

        public void setIdInt2(Integer idInt2)
        {
            this.idInt2 = idInt2;
        }

        public String getIdStr()
        {
            return idStr;
        }

        public void setIdStr(String idStr)
        {
            this.idStr = idStr;
        }

        public String getTitle()
        {
            return title;
        }

        public void setTitle(String title)
        {
            this.title = title;
        }

        public String getDescription()
        {
            return description;
        }

        public void setDescription(String description)
        {
            this.description = description;
        }

        public String toString()
        {
            return ToStringBuilder.reflectionToString(this).toString();
        }
    }

    public static class Role
    {
        private Integer actorId;
        private Integer actorId2;
        private Integer movieIntId;
        private Integer movieIntId2;
        private String movieStrId;

        public Role()
        {
        }

        public Integer getActorId()
        {
            return actorId;
        }

        public void setActorId(Integer actorId)
        {
            this.actorId = actorId;
        }

        public Integer getMovieIntId()
        {
            return movieIntId;
        }

        public Integer getMovieIntId2()
        {
            return movieIntId2;
        }

        public void setMovieIntId2(Integer movieIntId2)
        {
            this.movieIntId2 = movieIntId2;
        }

        public Integer getActorId2()
        {
            return actorId2;
        }

        public void setActorId2(Integer actorId2)
        {
            this.actorId2 = actorId2;
        }

        public void setMovieIntId(Integer movieIntId)
        {
            this.movieIntId = movieIntId;
        }

        public String getMovieStrId()
        {
            return movieStrId;
        }

        public void setMovieStrId(String movieStrId)
        {
            this.movieStrId = movieStrId;
        }

        public String toString()
        {
            return ToStringBuilder.reflectionToString(this).toString();
        }
    }
}
