package org.apache.ojb.broker;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.apache.ojb.broker.query.Criteria;
import org.apache.ojb.broker.query.QueryByCriteria;
import org.apache.ojb.junit.JUnitExtensions;

/**
 * Tests multithreaded read of objects using proxy for nested 1:1 references
 * Account --> Buyer --> Address --> AddressType
 *
 * @author <a href="mailto:armin@codeAuLait.de">Armin Waibel</a>
 * @version $Id: MultithreadedReadTest.java,v 1.3 2004/04/05 17:11:48 arminw Exp $
 */
public class MultithreadedReadTest extends JUnitExtensions.MultiThreadedTestCase
{
    int loops = 2;
    int concurrentThreads = 19;
    int numberOfObjects = 30;

    public MultithreadedReadTest(String s)
    {
        super(s);
    }

    public static void main(String[] args)
    {
        String[] arr = {MultithreadedReadTest.class.getName()};
        junit.textui.TestRunner.main(arr);
    }

    protected void setUp() throws Exception
    {
        super.setUp();
    }

    protected void tearDown() throws Exception
    {
        super.tearDown();
    }

    public void testClosedPB() throws Throwable
    {
        String name = "testClosedPB_"+System.currentTimeMillis();

        Integer[] ids = prepareTestRead(name, 5);
        Account account = null;
        PersistenceBroker broker = null;
        try
        {
            broker = PersistenceBrokerFactory.defaultPersistenceBroker();
            Criteria crit = new Criteria();
            crit.addIn("id", Arrays.asList(ids));
            QueryByCriteria query = new QueryByCriteria(Account.class, crit);
            Collection result = broker.getCollectionByQuery(query);
            Iterator iter = result.iterator();
            //			 iter.next();
            account = (Account) iter.next();
            while (iter.hasNext())
            {
                iter.next();
            }
        }
        finally
        {
            if (broker != null) broker.close();
        }

        TestCaseRunnable tct [] = new TestCaseRunnable[50];
        for (int i = 0; i < concurrentThreads; i++)
        {
            tct[i] = new TestHandleMaterialize(account, name);
        }
        // run test classes
        runTestCaseRunnables(tct);
    }

    /**
     * Read objects using lazy materialization for references from DB. Different threads
     * call the references on the read objects
     */
    public void testObjectMaterializationByDifferentThread() throws Exception
    {
        for (int k = 0; k < loops; k++)
        {
            String searchCriteria = "testObjectMaterializationByDifferentThread_" + System.currentTimeMillis();
            prepareTestRead(searchCriteria, concurrentThreads);

            Collection accounts;
            PersistenceBroker broker = null;
            try
            {
                broker = PersistenceBrokerFactory.defaultPersistenceBroker();
                Criteria crit = new Criteria();
                crit.addEqualTo("name", searchCriteria);
                QueryByCriteria query = new QueryByCriteria(Account.class, crit);
                accounts = broker.getCollectionByQuery(query);
                assertEquals(concurrentThreads, accounts.size());
            }
            finally
            {
                if (broker != null) broker.close();
            }
            Iterator iter = accounts.iterator();
            TestCaseRunnable tct [] = new TestCaseRunnable[concurrentThreads];
            for (int i = 0; i < concurrentThreads; i++)
            {
                tct[i] = new TestHandleMaterialize((Account) iter.next(), searchCriteria);
            }
            // run test classes
            runTestCaseRunnables(tct);
        }
    }

    /**
     * Different threads try to materialize the same bunch of objects.
     */
    public void testMultithreadedRead() throws Exception
    {
        String searchCriteria = "testMultithreadedRead_" + System.currentTimeMillis();
        prepareTestRead(searchCriteria, numberOfObjects);
        System.out.println("Multithreaded read of objects - start");
        System.out.println("" + concurrentThreads + " concurrent threads read "
                + numberOfObjects + " objects per thread, loop " + loops + " times");
        for (int k = 0; k < loops; k++)
        {
            PersistenceBroker broker = PersistenceBrokerFactory.defaultPersistenceBroker();
            broker.clearCache();
            broker.close();
            TestCaseRunnable tct [] = new TestCaseRunnable[concurrentThreads];
            for (int i = 0; i < concurrentThreads; i++)
            {
                tct[i] = new TestHandleRead(searchCriteria);
            }
            // run test classes
            runTestCaseRunnables(tct);
        }
        System.out.println();
        System.out.println("Multithreaded read of objects - end");
    }

    /**
     * Different threads try to materialize the same bunch of objects
     */
    public void testMultithreadedLazyRead() throws Exception
    {
        String name = "testMultithreadedLazyRead" + System.currentTimeMillis();
        List identityList = prepareTestLazyRead(name, concurrentThreads);
        System.out.println("Multithreaded lazy read of objects - start");
        System.out.println("" + concurrentThreads + " concurrent threads read different object with lazy" +
                " materialization reference, loop " + loops + " times");
        for (int k = 0; k < loops; k++)
        {
            PersistenceBroker broker = PersistenceBrokerFactory.defaultPersistenceBroker();
            broker.clearCache();
            broker.close();
            TestCaseRunnable tct [] = new TestCaseRunnable[concurrentThreads];
            for (int i = 0; i < concurrentThreads; i++)
            {
                tct[i] = new TestHandleLazyRead(identityList, name);
            }
            // run test classes
            runTestCaseRunnables(tct);
        }
        System.out.println();
        System.out.println("Multithreaded lazy read of objects - end");
    }

    private Integer[] prepareTestRead(String name, int numbers) throws Exception
    {
        Integer[] ids = new Integer[numbers];
        PersistenceBroker broker = PersistenceBrokerFactory.defaultPersistenceBroker();
        try
        {
            broker.beginTransaction();
            for (int i = 0; i < numbers; i++)
            {
                AddressType type = new AddressTypeImpl(name);
                Address address = new AddressImpl(name, type);
                Buyer buyer = new BuyerImpl(name, address);
                Account account = new AccountImpl(name, buyer);
                broker.store(account);
                ids[i] = account.getId();
            }
            broker.commitTransaction();
            broker.clearCache();
        }
        finally
        {
            if (broker != null) broker.close();
        }
        return ids;
    }

    private List prepareTestLazyRead(String searchCriteria, int numbers) throws Exception
    {
        List result = new ArrayList();
        PersistenceBroker broker = PersistenceBrokerFactory.defaultPersistenceBroker();
        try
        {
            broker.beginTransaction();
            for (int i = 0; i < numbers; i++)
            {
                AddressType type = new AddressTypeImpl(searchCriteria);
                Address address = new AddressImpl(searchCriteria, type);
                Buyer buyer = new BuyerImpl(searchCriteria, address);
                Account account = new AccountImpl(searchCriteria, buyer);
                broker.store(account);
                Identity oid = new Identity(account, broker);
                result.add(oid);
            }
            broker.commitTransaction();
            broker.clearCache();
        }
        finally
        {
            if (broker != null) broker.close();
        }
        return result;
    }


    //***********************************************
    // test handle of multithreaded test
    //***********************************************
    class TestHandleRead extends JUnitExtensions.MultiThreadedTestCase.TestCaseRunnable
    {
        String searchCriteria;

        public TestHandleRead(String searchCriteria)
        {
            this.searchCriteria = searchCriteria;
        }

        public void runTestCase() throws Throwable
        {
            readByCollection();
            readByIterator();
        }

        private void readByCollection() throws Exception
        {
            PersistenceBroker broker = null;
            try
            {
                broker = PersistenceBrokerFactory.defaultPersistenceBroker();
                broker.clearCache();
                Criteria crit = new Criteria();
                crit.addEqualTo("name", searchCriteria);
                QueryByCriteria query = new QueryByCriteria(Account.class, crit);
                Collection accounts = broker.getCollectionByQuery(query);
                assertEquals("Wrong number of expected objects", numberOfObjects, accounts.size());
                for (Iterator iter = accounts.iterator(); iter.hasNext();)
                {
                    Account account = (Account) iter.next();
                    assertEquals(searchCriteria, account.getName());
                    assertNotNull("All accounts have a reference to an Buyer", account.getBuyer());
                    assertNotNull("All buyers have a reference to an Address", account.getBuyer().getAddress());
                    assertNotNull("All addresses have a reference to an AdressType", account.getBuyer().getAddress().getType());
                    assertNotNull("All AddressType have a name", account.getBuyer().getAddress().getType().getName());
                    // System.out.println(""+Thread.currentThread().toString()+": passed");
                }
            }
            finally
            {
                if (broker != null) broker.close();
            }
        }

        private void readByIterator() throws Exception
        {
            PersistenceBroker broker = null;
            try
            {
                broker = PersistenceBrokerFactory.defaultPersistenceBroker();
                broker.clearCache();
                Criteria crit = new Criteria();
                crit.addEqualTo("name", searchCriteria);
                QueryByCriteria query = new QueryByCriteria(Account.class, crit);
                Iterator iter = broker.getIteratorByQuery(query);
                for (; iter.hasNext();)
                {
                    Account account = (Account) iter.next();
                    assertEquals(searchCriteria, account.getName());
                    assertNotNull("All accounts have a reference to an Buyer", account.getBuyer());
                    assertNotNull("All buyers have a reference to an Address", account.getBuyer().getAddress());
                    assertNotNull("All addresses have a reference to an AdressType", account.getBuyer().getAddress().getType());
                    assertNotNull("All AddressType have a name", account.getBuyer().getAddress().getType().getName());
                    // System.out.println(""+Thread.currentThread().toString()+": passed");
                }
            }
            finally
            {
                if (broker != null) broker.close();
            }
        }
    }

    class TestHandleLazyRead extends JUnitExtensions.MultiThreadedTestCase.TestCaseRunnable
    {
        List identityList;
        String name;

        public TestHandleLazyRead(List identityList, String name)
        {
            this.identityList = identityList;
            this.name = name;
        }

        public void runTestCase() throws Throwable
        {
            PersistenceBroker broker = null;
            Account account = null;
            try
            {
                broker = PersistenceBrokerFactory.defaultPersistenceBroker();
                Iterator it = identityList.iterator();
                while (it.hasNext())
                {
                    Identity oid = (Identity) it.next();
                    account = (Account) broker.getObjectByIdentity(oid);
                }
            }
            finally
            {
                if (broker != null) broker.close();
            }
            assertEquals(name, account.getName());
            assertNotNull("All accounts have a reference to an Buyer", account.getBuyer());
            assertNotNull("All buyers have a reference to an Address", account.getBuyer().getAddress());
            assertNotNull("All addresses have a reference to an AdressType", account.getBuyer().getAddress().getType());
            assertNotNull("All AddressType have a name", account.getBuyer().getAddress().getType().getName());
        }
    }


    class TestHandleMaterialize extends JUnitExtensions.MultiThreadedTestCase.TestCaseRunnable
    {
        Account account;
        String name;

        public TestHandleMaterialize(Account account, String name)
        {
            this.account = account;
            this.name = name;
        }

        public void runTestCase() throws Throwable
        {
            assertEquals(name, account.getName());
            assertNotNull("All accounts have a reference to an Buyer", account.getBuyer());
            assertEquals(name, account.getBuyer().getName());
            assertNotNull("All buyers have a reference to an Address", account.getBuyer().getAddress());
            assertEquals(name, account.getBuyer().getAddress().getName());
            assertNotNull("All addresses have a reference to an AdressType", account.getBuyer().getAddress().getType());
            assertNotNull("All AddressType have a name", account.getBuyer().getAddress().getType().getName());
        }
    }


    //***********************************************
    // test classes/interfaces starts here
    //***********************************************
    public interface Account extends Base
    {
        Buyer getBuyer();

        void setBuyer(Buyer buyer);
    }

    public static class AccountImpl extends BaseImpl implements Account
    {
        Buyer buyer;

        public AccountImpl(String name, Buyer buyer)
        {
            super(name);
            this.buyer = buyer;
        }

        public AccountImpl(Buyer buyer)
        {
            this.buyer = buyer;
        }

        public AccountImpl()
        {

        }

        public Buyer getBuyer()
        {
            return buyer;
        }

        public void setBuyer(Buyer buyer)
        {
            this.buyer = buyer;
        }
    }

    public interface Buyer extends Base
    {
        Address getAddress();

        void setAddress(Address address);
    }

    public static class BuyerImpl extends BaseImpl implements Buyer
    {
        Address address;

        public BuyerImpl(String name, Address address)
        {
            super(name);
            this.address = address;
        }

        public BuyerImpl(Address address)
        {
            this.address = address;
        }

        public BuyerImpl()
        {

        }

        public Address getAddress()
        {
            return address;
        }

        public void setAddress(Address address)
        {
            this.address = address;
        }
    }

    public interface Address extends Base
    {
        AddressType getType();

        void setType(AddressType type);
    }

    public static class AddressImpl extends BaseImpl implements Address
    {
        AddressType type;

        public AddressImpl(String name, AddressType type)
        {
            super(name);
            this.type = type;
        }

        public AddressImpl(AddressType type)
        {
            this.type = type;
        }

        public AddressImpl()
        {

        }

        public AddressType getType()
        {
            return type;
        }

        public void setType(AddressType type)
        {
            this.type = type;
        }
    }

    public interface AddressType extends Base
    {
    }

    public static class AddressTypeImpl extends BaseImpl implements AddressType
    {
        public AddressTypeImpl(String name)
        {
            super(name);
        }

        public AddressTypeImpl()
        {
        }
    }

    public interface Base
    {
        Integer getId();

        void setId(Integer id);

        String getName();

        void setName(String name);
    }

    public static class BaseImpl
    {
        Integer id;
        String name;

        public BaseImpl(String name)
        {
            this.name = name;
        }

        public BaseImpl()
        {
        }

        public Integer getId()
        {
            return id;
        }

        public void setId(Integer id)
        {
            this.id = id;
        }

        public String getName()
        {
            return name;
        }

        public void setName(String name)
        {
            this.name = name;
        }
    }
}
