package org.apache.ojb.broker.sequence;

import junit.framework.TestCase;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.ojb.broker.Identity;
import org.apache.ojb.broker.PersistenceBroker;
import org.apache.ojb.broker.PersistenceBrokerFactory;
import org.apache.ojb.broker.TestHelper;
import org.apache.ojb.broker.metadata.MetadataManager;
import org.apache.ojb.broker.metadata.SequenceDescriptor;
import org.apache.ojb.broker.platforms.PlatformHsqldbImpl;
import org.apache.ojb.broker.platforms.PlatformMySQLImpl;
import org.apache.ojb.broker.platforms.Platform;
import org.apache.ojb.broker.query.Criteria;
import org.apache.ojb.broker.query.QueryByCriteria;
import org.apache.ojb.broker.query.QueryFactory;
import org.apache.ojb.broker.util.sequence.SequenceManagerNativeImpl;
import org.apache.ojb.odmg.OJB;
import org.odmg.Implementation;
import org.odmg.Database;
import org.odmg.Transaction;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Collection;

/**
 * Test case for {@link SequenceManagerNativeImpl}. These test check
 * support for native identity columns. Test case only works for
 * Hsql and Mysql.
 *
 * @author <a href="mailto:armin@codeAuLait.de">Armin Waibel</a>
 * @version $Id: NativeIdentifierTest.java,v 1.10 2004/05/11 11:55:57 arminw Exp $
 */
public class NativeIdentifierTest extends TestCase
{
    // Statements for MainObject table
    private static final String DROP = "DROP TABLE NATIVE_MAIN_OBJECT";
    private static final String CREATE_MYSQL =
            "CREATE TABLE NATIVE_MAIN_OBJECT(NATIVE_ID int(11) NOT NULL PRIMARY KEY" +
            " auto_increment,REF_ID int(11),NAME VARCHAR(250))";
    private static final String CREATE_HSQL =
            "CREATE TABLE NATIVE_MAIN_OBJECT(NATIVE_ID IDENTITY NOT NULL PRIMARY KEY," +
            " REF_ID int(11),NAME VARCHAR(250))";
    private static final String INSERT_DUMMY = "INSERT INTO NATIVE_MAIN_OBJECT (NAME) VALUES ('Dummy_1')";


    // Statements for NATIVE_REF_TEST table
    private static final String CREATE_REF_MYSQL =
            "CREATE TABLE NATIVE_REFERENCE_OBJECT (NATIVE_ID int(11) NOT NULL PRIMARY KEY auto_increment," +
            " NAME VARCHAR(250), FK_ID BIGINT, REF_ID int(11), SINGLE_REF_FK BIGINT)";
    private static final String CREATE_REF_HSQL =
            "CREATE TABLE NATIVE_REFERENCE_OBJECT (NATIVE_ID IDENTITY NOT NULL PRIMARY KEY," +
            " NAME VARCHAR(250), FK_ID BIGINT, REF_ID int(11), SINGLE_REF_FK BIGINT)";
    private static final String DROP_REF = "DROP TABLE NATIVE_REFERENCE_OBJECT";
    private static final String INSERT_DUMMY_REF = "INSERT INTO NATIVE_REFERENCE_OBJECT (NAME) VALUES ('Dummy_2')";


    private PersistenceBroker broker;
    private Platform platform;
    private Class oldSequenceManager;

    public NativeIdentifierTest(String s)
    {
        super(s);
    }

    public static void main(String[] args)
    {
        String[] arr = {NativeIdentifierTest.class.getName()};
        junit.textui.TestRunner.main(arr);
    }

    private boolean skipTest() throws Exception
    {
        boolean skip = !((platform instanceof PlatformMySQLImpl) || (platform instanceof PlatformHsqldbImpl));
        return skip;
    }

    protected void setUp() throws Exception
    {
        super.setUp();


        PersistenceBroker pb = PersistenceBrokerFactory.defaultPersistenceBroker();
        platform = pb.serviceConnectionManager().getSupportedPlatform();
        if (skipTest()) return;
        Connection con;
        Statement stmt;

        try
        {
            con = pb.serviceConnectionManager().getConnection();
            stmt = con.createStatement();
            if(platform instanceof PlatformMySQLImpl)
            {
                stmt.execute(CREATE_MYSQL);
            }
            if(platform instanceof PlatformHsqldbImpl)
            {
                stmt.execute(CREATE_HSQL);
            }
            stmt.close();

            stmt = con.createStatement();
            stmt.execute(INSERT_DUMMY);
            stmt.close();

            stmt = con.createStatement();
            if(platform instanceof PlatformMySQLImpl)
            {
                stmt.execute(CREATE_REF_MYSQL);
            }
            if(platform instanceof PlatformHsqldbImpl)
            {
                stmt.execute(CREATE_REF_HSQL);
            }
            stmt.close();

            stmt = con.createStatement();
            stmt.execute(INSERT_DUMMY_REF);
            stmt.close();

            SequenceDescriptor sd = MetadataManager.getInstance().connectionRepository().
                    getDescriptor(pb.getPBKey()).getSequenceDescriptor();
            oldSequenceManager = sd.getSequenceManagerClass();
            sd.setSequenceManagerClass(SequenceManagerNativeImpl.class);
        }
        finally
        {
            if (pb != null) pb.close();
        }

        PersistenceBrokerFactory.releaseAllInstances();

        broker = PersistenceBrokerFactory.defaultPersistenceBroker();
        SequenceDescriptor sd = MetadataManager.getInstance().connectionRepository().
                getDescriptor(broker.getPBKey()).getSequenceDescriptor();
        assertEquals(SequenceManagerNativeImpl.class, sd.getSequenceManagerClass());
    }

    protected void tearDown() throws Exception
    {
        super.tearDown();

        try
        {
            broker.close();
        }
        catch (Exception e)
        {
        }
        if (skipTest()) return;

        Connection con;
        Statement stmt;
        PersistenceBroker pb = PersistenceBrokerFactory.defaultPersistenceBroker();
        try
        {
            con = pb.serviceConnectionManager().getConnection();
            stmt = con.createStatement();
            stmt.execute(DROP);
            stmt.close();

            stmt = con.createStatement();
            stmt.execute(DROP_REF);
            stmt.close();

            SequenceDescriptor sd = MetadataManager.getInstance().connectionRepository().
                    getDescriptor(pb.getPBKey()).getSequenceDescriptor();
            sd.setSequenceManagerClass(oldSequenceManager);
        }
        finally
        {
            if (pb != null) pb.close();
        }

        PersistenceBrokerFactory.releaseAllInstances();
        broker = PersistenceBrokerFactory.defaultPersistenceBroker();
        SequenceDescriptor sd = MetadataManager.getInstance().connectionRepository().
                getDescriptor(broker.getPBKey()).getSequenceDescriptor();
        assertEquals(oldSequenceManager, sd.getSequenceManagerClass());
        broker.close();
    }

    public void testSimpleInsert() throws Exception
    {
        long timestamp = (long) (System.currentTimeMillis() * Math.random());
        String name = "testSimpleInsert_" + timestamp;

        if (skipTest()) return;

        MainObject obj_1 = new MainObject(null, name);
        MainObject obj_2 = new MainObject(null, name);
        MainObject obj_3 = new MainObject(null, name);

        broker.beginTransaction();
        broker.store(obj_1);
//        System.out.println("obj_1: "+obj_1);
        broker.store(obj_2);
//        System.out.println("obj_2: "+obj_2);
        broker.store(obj_3);
//        System.out.println("obj_3: "+obj_3);
        broker.commitTransaction();

        Criteria crit = new Criteria();
        crit.addEqualTo("name", name);
        QueryByCriteria query = QueryFactory.newQuery(MainObject.class, crit);
        int result = broker.getCount(query);
        assertEquals("Not all objects created", 3, result);
        assertNotNull(obj_1.getIdentifier());
    }

    public void testSimpleInsertODMG() throws Exception
    {
        long timestamp = (long) (System.currentTimeMillis() * Math.random());
        String name = "testSimpleInsert_" + timestamp;

        if (skipTest()) return;

        MainObject obj_1 = new MainObject(null, name);
        MainObject obj_2 = new MainObject(null, name);
        MainObject obj_3 = new MainObject(null, name);

        Implementation odmg = OJB.getInstance();
        Database db = odmg.newDatabase();
        db.open(TestHelper.DEF_DATABASE_NAME, Database.OPEN_READ_WRITE);

        Transaction tx = odmg.newTransaction();
        tx.begin();
        tx.lock(obj_1, Transaction.WRITE);
        tx.lock(obj_2, Transaction.WRITE);
        tx.lock(obj_3, Transaction.WRITE);
        tx.commit();

        Criteria crit = new Criteria();
        crit.addEqualTo("name", name);
        QueryByCriteria query = QueryFactory.newQuery(MainObject.class, crit);
        int result = broker.getCount(query);
        assertEquals("Not all objects created", 3, result);
        assertNotNull(obj_1.getIdentifier());
    }

    public void testAllInOne() throws Exception
    {
        testSimpleInsert();
        testSimpleInsert();
        testReferenceInsertUpdate();
        testReferenceInsertUpdate();
    }

    public void testReferenceInsertUpdate() throws Exception
    {
        if (skipTest()) return;
        long timestamp = (long) (System.currentTimeMillis() * Math.random());
        String name = "testReferenceInsert_main_" + timestamp;
        String nameRef = "testReferenceInsert_reference_" + timestamp;
        String nameSingleRef = "testReferenceInsert_single_reference_" + timestamp;

        MainObject obj_1 = new MainObject(null, name);
        MainObject obj_2 = new MainObject(null, name);

        SingleReference s_ref_1 = new SingleReference(nameSingleRef);
        SingleReference s_ref_2 = new SingleReference(nameSingleRef);

        CollectionReference ref_1 = new CollectionReference(null, nameRef);
        CollectionReference ref_2 = new CollectionReference(null, nameRef);
        CollectionReference ref_3 = new CollectionReference(null, nameRef);
        CollectionReference ref_4 = new CollectionReference(null, nameRef);
        ref_1.setSingleReference(s_ref_1);
        ref_4.setSingleReference(s_ref_2);

        SingleReference s_ref_3 = new SingleReference(nameSingleRef);
        SingleReference s_ref_4 = new SingleReference(nameSingleRef);

        obj_1.addReference(ref_1);
        obj_1.addReference(ref_2);
        obj_1.addReference(ref_3);
        obj_1.addReference(ref_4);

        obj_1.setSingleReference(s_ref_3);
        s_ref_3.setMainObject(obj_1);
        obj_2.setSingleReference(s_ref_4);
        s_ref_3.setMainObject(obj_1);

        broker.beginTransaction();
        // first store a reference
        broker.store(ref_1);
//        System.out.println("ref_1: "+ref_1);
        // then store main object with other references
        broker.store(obj_1);
//        System.out.println("obj_1: "+obj_1);
        // store second object without references
        broker.store(obj_2);
//        System.out.println("obj_2: "+obj_2);
        broker.commitTransaction();

        // try to find both objects
        Criteria crit = new Criteria();
        crit.addEqualTo("name", name);
        QueryByCriteria query = QueryFactory.newQuery(MainObject.class, crit);
        int result = broker.getCount(query);
        assertEquals("Wrong object count", 2, result);

        // pk have to set and have to be different
        assertNotNull(obj_1.getIdentifier());
        assertNotNull(obj_2.getIdentifier());
        assertNotSame(obj_1.getIdentifier(), obj_2.getIdentifier());
        assertTrue(obj_1.getIdentifier().longValue() > 0);
        assertTrue(obj_2.getIdentifier().longValue() > 0);

        // get Identity objects
        Identity oid_1 = new Identity(obj_1, broker);
        Identity oid_2 = new Identity(obj_2, broker);
        // get identifier (PK) values
        Long id_1 = obj_1.getIdentifier();
        Long id_2 = obj_2.getIdentifier();

        broker.clearCache();

        // get object with references
        obj_1 = (MainObject) broker.getObjectByIdentity(oid_1);
        assertNotNull(obj_1);
        List references = obj_1.getAllReferences();
        assertNotNull(references);
        assertEquals("4 references expected for object: "+obj_1, 4, references.size());
        Iterator it = references.iterator();
        while (it.hasNext())
        {
            CollectionReference ref = (CollectionReference) it.next();
            assertEquals("Main object fk expected", obj_1.getIdentifier(), ref.fkIdentifier);
            assertTrue("We expect a positive value, identity columns have to start > 0",
                    (ref.getRefIdentifier().longValue() > 0));
        }
        assertNotNull(obj_1.getSingleReference());
        obj_2 = (MainObject) broker.getObjectByIdentity(oid_2);
        assertTrue(obj_1.getIdentifier().longValue() > 0);
        assertTrue(obj_2.getIdentifier().longValue() > 0);
        assertNotNull(obj_2.getSingleReference());
        assertTrue(obj_2.getSingleReference().getId().longValue() > 0);
        assertTrue(obj_1.getSingleReference().getId().longValue() > 0);
        assertNotSame(obj_1.getSingleReference(), obj_2.getSingleReference());
        broker.clearCache();

        // get references only
        Criteria crit_2 = new Criteria();
        crit_2.addEqualTo("refName", nameRef);
        QueryByCriteria query_2 = QueryFactory.newQuery(CollectionReference.class, crit_2);
        int result_2 = broker.getCount(query_2);
        assertEquals("Not all objects created", 4, result_2);
        assertNotNull(ref_3.getRefIdentifier());

        broker.clearCache();

        // get second object
        MainObject retObj = (MainObject) broker.getObjectByIdentity(oid_2);
        List refList = retObj.getAllReferences();
        assertNotNull(refList);
        assertEquals("object do not have references", 0, refList.size());

        // add new reference to object
        CollectionReference ref_5 = new CollectionReference(null, nameRef);
        CollectionReference ref_6 = new CollectionReference(null, nameRef);
        obj_1.addReference(ref_5);
        obj_2.addReference(ref_6);
        broker.beginTransaction();
        broker.store(obj_1);
        broker.store(obj_2);
        broker.commitTransaction();
        assertNotNull(ref_5.getRefIdentifier());
        assertNotNull(ref_6.getRefIdentifier());
        assertEquals(id_1, obj_1.getIdentifier());
        assertEquals(id_2, obj_2.getIdentifier());

        obj_1 = (MainObject) broker.getObjectByIdentity(oid_1);
        assertNotNull(obj_1);
        references = obj_1.getAllReferences();
        assertNotNull(references);
        assertEquals("5 references expected for object: "+obj_1, 5, references.size());

        obj_2 = (MainObject) broker.getObjectByIdentity(oid_2);
        assertNotNull(obj_2);
        references = obj_2.getAllReferences();
        assertNotNull(references);
        assertEquals("1 references expected for object: "+obj_2, 1, references.size());

        assertEquals(id_1, obj_1.getIdentifier());
        assertEquals(id_2, obj_2.getIdentifier());

        // now update main objects
        obj_1.setName(name+"_update");
        obj_2.setName(name+"_update");
        broker.beginTransaction();
        broker.store(obj_1);
        broker.store(obj_2);
        broker.commitTransaction();

        obj_1 = (MainObject) broker.getObjectByIdentity(oid_1);
        obj_2 = (MainObject) broker.getObjectByIdentity(oid_2);

        assertNotNull(obj_1);
        assertNotNull(obj_2);
        assertEquals(obj_1.getName(), name+"_update");
        assertEquals(obj_2.getName(), name+"_update");
        assertEquals(id_1, obj_1.getIdentifier());
        assertEquals(id_2, obj_2.getIdentifier());

        // now update reference
        obj_2 = (MainObject) broker.getObjectByIdentity(oid_2);
        assertNotNull(obj_2);
        references = obj_2.getAllReferences();
        CollectionReference ref = (CollectionReference) references.get(0);
        ref.setRefName(nameRef+"_update");
        broker.beginTransaction();
        broker.store(obj_2);
        broker.commitTransaction();
        obj_2 = (MainObject) broker.getObjectByIdentity(oid_2);
        assertNotNull(obj_2);
        references = obj_2.getAllReferences();
        ref = (CollectionReference) references.get(0);
        assertEquals(nameRef+"_update", ref.getRefName());
        assertEquals(id_1, obj_1.getIdentifier());
        assertEquals(id_2, obj_2.getIdentifier());
    }

    public void testDelete() throws Exception
    {
        if (skipTest()) return;
        long timestamp = (long) (System.currentTimeMillis() * Math.random());
        String name = "testDelete_main_" + timestamp;
        String nameRef = "testDelete_reference_" + timestamp;

        MainObject obj_1 = new MainObject(null, name);

        CollectionReference ref_1 = new CollectionReference(null, nameRef);
        CollectionReference ref_2 = new CollectionReference(null, nameRef);

        obj_1.addReference(ref_1);
        obj_1.addReference(ref_2);
        broker.beginTransaction();
        broker.store(obj_1);
        broker.commitTransaction();
        Identity oid_1 = new Identity(obj_1, broker);

        MainObject result = (MainObject) broker.getObjectByIdentity(oid_1);
        assertNotNull(result);
        assertNotNull(result.getAllReferences());
        assertEquals(2, result.getAllReferences().size());
        Long fk = ((CollectionReference) result.getAllReferences().get(0)).getFkIdentifier();
        assertNotNull(result.getIdentifier());
        assertEquals(result.getIdentifier(), fk);

        broker.beginTransaction();
        broker.delete(obj_1);
        broker.commitTransaction();

        result = (MainObject) broker.getObjectByIdentity(oid_1);
        assertNull(result);
        Criteria crit_2 = new Criteria();
        crit_2.addEqualTo("refName", nameRef);
        QueryByCriteria query_2 = QueryFactory.newQuery(CollectionReference.class, crit_2);
        int result_2 = broker.getCount(query_2);
        assertEquals(0, result_2);
    }

    public void testDelete_2() throws Exception
    {
        if (skipTest()) return;
        long timestamp = (long) (System.currentTimeMillis() * Math.random());
        String name = "testDelete_2_main_" + timestamp;
        String nameRef = "testDelete_2_reference_" + timestamp;

        MainObject obj_1 = new MainObject(null, name);

        CollectionReference ref_1 = new CollectionReference(null, nameRef);
        CollectionReference ref_2 = new CollectionReference(null, nameRef);

        obj_1.addReference(ref_1);
        obj_1.addReference(ref_2);

        // chaotic operations
        broker.beginTransaction();
        // System.out.println("1. "+obj_1);
        broker.store(obj_1);
        // System.out.println("2. "+obj_1);
        broker.delete(obj_1);
        // System.out.println("3. "+obj_1);
        broker.store(obj_1);
        // System.out.println("4. "+obj_1);
        broker.delete(obj_1);
        // System.out.println("5. "+obj_1);
        broker.store(obj_1);
        // System.out.println("6. "+obj_1);
        broker.delete(obj_1);
        // System.out.println("7. "+obj_1);
        broker.store(obj_1);
        // System.out.println("8. "+obj_1);
        broker.commitTransaction();
        Identity oid_1 = new Identity(obj_1, broker);

        MainObject result = (MainObject) broker.getObjectByIdentity(oid_1);
        assertNotNull(result);
        assertNotNull(result.getAllReferences());
        assertEquals(2, result.getAllReferences().size());
        Long fk = ((CollectionReference) result.getAllReferences().get(0)).getFkIdentifier();
        assertNotNull(result.getIdentifier());
        assertEquals(result.getIdentifier(), fk);

        // we should find exactly one object
        Criteria c = new Criteria();
        c.addEqualTo("name", name);
        QueryByCriteria q = QueryFactory.newQuery(MainObject.class, c);
        Collection col = broker.getCollectionByQuery(q);
        assertNotNull(col);
        assertEquals(1, col.size());

        broker.beginTransaction();
        broker.delete(obj_1);
        broker.commitTransaction();

        result = (MainObject) broker.getObjectByIdentity(oid_1);
        assertNull(result);
        Criteria crit_2 = new Criteria();
        crit_2.addEqualTo("refName", nameRef);
        QueryByCriteria query_2 = QueryFactory.newQuery(CollectionReference.class, crit_2);
        int result_2 = broker.getCount(query_2);
        // System.out.println("## result: "+broker.getCollectionByQuery(query_2));
        assertEquals(0, result_2);
    }



    public static class MainObject implements Serializable
    {
        private Long identifier;
        private String name;
        private List allReferences;
        // we use anonymous field for FK
        private SingleReference singleReference;


        public MainObject()
        {
        }

        public MainObject(Long identifier, String name)
        {
            this.identifier = identifier;
            this.name = name;
        }

        public SingleReference getSingleReference()
        {
            return singleReference;
        }

        public void setSingleReference(SingleReference singleReference)
        {
            this.singleReference = singleReference;
        }

        public List getAllReferences()
        {
            return allReferences;
        }

        public void addReference(CollectionReference reference)
        {
            if (allReferences == null)
            {
                allReferences = new ArrayList();
            }
            allReferences.add(reference);
        }

        public void setAllReferences(List allReferences)
        {
            this.allReferences = allReferences;
        }

        public Long getIdentifier()
        {
            return identifier;
        }

        public void setIdentifier(Long identifier)
        {
            this.identifier = identifier;
        }

        public String getName()
        {
            return name;
        }

        public void setName(String name)
        {
            this.name = name;
        }

        public String toString()
        {
            return ToStringBuilder.reflectionToString(this);
        }
    }

    public static class SingleReference
    {
        Long id;
        String name;
        MainObject mainObject;

        public SingleReference()
        {
            this(null);
        }

        public SingleReference(String name)
        {
            this.name = name;
            // id = new Long((long)(Math.random() * Integer.MAX_VALUE));
        }

        public MainObject getMainObject()
        {
            return mainObject;
        }

        public void setMainObject(MainObject mainObject)
        {
            this.mainObject = mainObject;
        }

        public Long getId()
        {
            return id;
        }

        public void setId(Long id)
        {
            this.id = id;
        }

        public String getName()
        {
            return name;
        }

        public void setName(String name)
        {
            this.name = name;
        }
    }

    public static class CollectionReference implements Serializable
    {
        private Long refIdentifier;
        private String refName;
        private Long fkIdentifier;
        private SingleReference singleReference;

        public CollectionReference()
        {
        }

        public CollectionReference(Long refIdentifier, String refName)
        {
            this.refIdentifier = refIdentifier;
            this.refName = refName;
        }

        public Long getRefIdentifier()
        {
            return refIdentifier;
        }

        public void setRefIdentifier(Long refIdentifier)
        {
            this.refIdentifier = refIdentifier;
        }

        public SingleReference getSingleReference()
        {
            return singleReference;
        }

        public void setSingleReference(SingleReference singleReference)
        {
            this.singleReference = singleReference;
        }

        public Long getFkIdentifier()
        {
            return fkIdentifier;
        }

        public void setFkIdentifier(Long fkIdentifier)
        {
            this.fkIdentifier = fkIdentifier;
        }

        public String getRefName()
        {
            return refName;
        }

        public void setRefName(String refName)
        {
            this.refName = refName;
        }

        public String toString()
        {
            return ToStringBuilder.reflectionToString(this);
        }
    }
}
