package org.apache.ojb.odmg;

/**
 * This is the TestSuite that checks the ODMG locking strategies.
 * It performs 17 testMethods as defined in the Locking documentation.
 *
 * @author Thomas Mahler
 */

import junit.framework.TestCase;
import org.apache.ojb.broker.TestHelper;
import org.apache.ojb.odmg.locking.LockStrategy;
import org.apache.ojb.odmg.locking.SerializableStrategy;
import org.apache.ojb.odmg.shared.Article;
import org.odmg.Database;
import org.odmg.Implementation;
import org.odmg.ODMGException;

public class LockTestSerializable extends TestCase
{
    TransactionImpl tx1;
    TransactionImpl tx2;
    String databaseName;
    Article obj;
    LockStrategy ls;

    public static void main(String[] args)
    {
        String[] arr = { CLASS.getName()};
        junit.textui.TestRunner.main(arr);
    }

    private static Class CLASS = LockTestSerializable.class;

    public LockTestSerializable(String name)

    {
        super(name);
        ls = new SerializableStrategy();
    }

    public void setUp()
    {
        try
        {
            databaseName = TestHelper.DEF_DATABASE_NAME;

            // get facade instance
            Implementation odmg = OJB.getInstance();
            Database db = odmg.newDatabase();
            //open database
            db.open(databaseName, Database.OPEN_READ_WRITE);

            tx1 = (TransactionImpl) odmg.newTransaction();
            tx2 = (TransactionImpl) odmg.newTransaction();
            tx1.begin();
            tx2.begin();
            obj = new Article();
            obj.setArticleId(333);

        }
        catch (ODMGException e)
        {
        }

    }

    /**
     * Test 1
     */
    public void testSingleReadLock()
    {
        assertTrue(ls.readLock(tx1, obj));
    }

    /**
     * Test 2
     */
    public void testUpgradeReadLock()
    {
        assertTrue(ls.readLock(tx1, obj));
        assertTrue(ls.upgradeLock(tx1, obj));
    }

    /**
     * Test3
     */
    public void testReadThenWrite()
    {
        assertTrue(ls.readLock(tx1, obj));
        assertTrue(ls.writeLock(tx1, obj));
    }

    /**
     * Test 4
     */
    public void testSingleWriteLock()
    {
        assertTrue(ls.writeLock(tx1, obj));
    }

    /**
     * Test 5
     */
    public void testWriteThenRead()
    {
        assertTrue(ls.writeLock(tx1, obj));
        assertTrue(ls.readLock(tx1, obj));
    }

    /**
     * Test 6
     */
    public void testMultipleReadLock()
    {
        assertTrue(ls.readLock(tx1, obj));
        assertTrue(!ls.readLock(tx2, obj));
    }

    /**
     * Test 7
     */
    public void testUpgradeWithExistingReader()
    {
        assertTrue(ls.readLock(tx1, obj));
        assertTrue(!ls.upgradeLock(tx2, obj));
    }

    /**
     * Test 8
     */
    public void testWriteWithExistingReader()
    {
        assertTrue(ls.readLock(tx1, obj));
        assertTrue(!ls.writeLock(tx2, obj));
    }

    /**
     * Test 9
     */
    public void testUpgradeWithMultipleReaders()
    {
        assertTrue(ls.readLock(tx1, obj));
        assertTrue(!ls.readLock(tx2, obj));
        assertTrue(!ls.upgradeLock(tx2, obj));
    }

    /**
     * Test 10
     */
    public void testWriteWithMultipleReaders()
    {
        assertTrue(ls.readLock(tx1, obj));
        assertTrue(!ls.readLock(tx2, obj));
        assertTrue(!ls.writeLock(tx2, obj));
    }

    /**
     * Test 11
     */
    public void testUpgradeWithMultipleReadersOn1()
    {
        assertTrue(ls.readLock(tx1, obj));
        assertTrue(!ls.readLock(tx2, obj));
        assertTrue(ls.upgradeLock(tx1, obj));
    }

    /**
     * Test 12
     */
    public void testWriteWithMultipleReadersOn1()
    {
        assertTrue(ls.readLock(tx1, obj));
        assertTrue(!ls.readLock(tx2, obj));
        assertTrue(ls.writeLock(tx1, obj));
    }

    /**
     * Test 13
     */
    public void testReadWithExistingWriter()
    {
        assertTrue(ls.writeLock(tx1, obj));
        assertTrue(!ls.readLock(tx2, obj));
    }

    /**
     * Test 14
     */
    public void testMultipleWriteLock()
    {
        assertTrue(ls.writeLock(tx1, obj));
        assertTrue(!ls.writeLock(tx2, obj));
    }

    /**
     * Test 15
     */
    public void testReleaseReadLock()
    {
        assertTrue(ls.readLock(tx1, obj));
        assertTrue(ls.releaseLock(tx1, obj));
        assertTrue(ls.writeLock(tx2, obj));
    }

    /**
     * Test 16
     */
    public void testReleaseUpgradeLock()
    {
        assertTrue(ls.upgradeLock(tx1, obj));
        assertTrue(ls.releaseLock(tx1, obj));
        assertTrue(ls.writeLock(tx2, obj));
    }

    /**
     * Test 17
     */
    public void testReleaseWriteLock()
    {
        assertTrue(ls.writeLock(tx1, obj));
        assertTrue(ls.releaseLock(tx1, obj));
        assertTrue(ls.writeLock(tx2, obj));
    }

    /**
     * Test 18
     */
    public void testReadThenRead()
    {
        assertTrue(ls.readLock(tx1, obj));
        assertTrue(ls.readLock(tx1, obj));
    }

    /**
     * Insert the method's description here.
     * Creation date: (06.12.2000 21:59:14)
     */
    public void tearDown()
    {
        try
        {
            ls.releaseLock(tx1, obj);
            ls.releaseLock(tx2, obj);
            tx1.abort();
            tx2.abort();

            OJB.getInstance().getDatabase(null).close();
            databaseName = null;
        }
        catch (Throwable t)
        {
        }
    }
}
