package org.apache.ojb.odmg;


import java.util.Collection;
import java.util.List;

import junit.framework.TestCase;
import org.apache.ojb.broker.LockedByVersion;
import org.apache.ojb.broker.PersistenceBroker;
import org.apache.ojb.broker.PersistenceBrokerFactory;
import org.apache.ojb.broker.TestHelper;
import org.apache.ojb.broker.query.Criteria;
import org.apache.ojb.broker.query.QueryFactory;
import org.apache.ojb.odmg.locking.LockManager;
import org.apache.ojb.odmg.locking.LockManagerFactory;
import org.apache.ojb.odmg.shared.Article;
import org.odmg.Database;
import org.odmg.Implementation;
import org.odmg.LockNotGrantedException;
import org.odmg.OQLQuery;
import org.odmg.Transaction;

/**
 * Test optimistic and pessimistic locking mechanisms
 */
public class LockingTest extends TestCase
{
    public static void main(String[] args)
    {
        String[] arr = {LockingTest.class.getName()};
        junit.textui.TestRunner.main(arr);
    }

    private static String PRE = "LockingTest_" + System.currentTimeMillis() + "_";

    private Implementation odmg1 = OJB.getInstance();
    private Database db1 = odmg1.newDatabase();

    private Implementation odmg2 = OJB.getInstance();
    private Database db2 = odmg2.newDatabase();

    /**
     * Insert the method's description here.
     * Creation date: (24.12.2000 00:33:40)
     */
    public LockingTest(String name)
    {
        super(name);
    }


    /**
     * Insert the method's description here.
     * Creation date: (06.12.2000 21:58:53)
     */
    public void setUp() throws Exception
    {
        String databaseName = TestHelper.DEF_DATABASE_NAME;

        odmg1 = OJB.getInstance();
        db1 = odmg1.newDatabase();
        db1.open(databaseName, Database.OPEN_READ_WRITE);

        odmg2 = OJB.getInstance();
        db2 = odmg2.newDatabase();
        db2.open(databaseName, Database.OPEN_READ_WRITE);

    }

    /**
     * Insert the method's description here.
     * Creation date: (06.12.2000 21:59:14)
     */
    public void tearDown() throws Exception
    {
        if (odmg1.currentTransaction() != null) odmg1.currentTransaction().abort();
        db1.close();

        if (odmg2.currentTransaction() != null) odmg2.currentTransaction().abort();
        db2.close();
    }

    /**
     * Test multiple locks on the same object
     */
    public void testMultipleLocks() throws Exception
    {
        long timestamp = System.currentTimeMillis();
        String name = "testMultipleLocks_" + timestamp;
        String nameUpdated = "testMultipleLocks_Updated_" + timestamp;
        TransactionExt tx = (TransactionExt) odmg1.newTransaction();
        LockedByVersion obj = new LockedByVersion();
        tx.begin();
        tx.lock(obj, Transaction.WRITE);
        obj.setValue(name);
        tx.lock(obj, Transaction.WRITE);
        tx.lock(obj, Transaction.UPGRADE);
        tx.commit();

        OQLQuery query = odmg1.newOQLQuery();
        query.create("select all from " + LockedByVersion.class.getName() + " where value like $1");
        query.bind(name);
        Collection result = (Collection) query.execute();
        assertNotNull(result);
        assertEquals(1, result.size());

        tx.begin();
        tx.lock(obj, Transaction.WRITE);
        tx.lock(obj, Transaction.WRITE);
        obj.setValue(nameUpdated);
        tx.lock(obj, Transaction.WRITE);
        tx.lock(obj, Transaction.UPGRADE);
        tx.commit();

        query = odmg1.newOQLQuery();
        query.create("select all from " + LockedByVersion.class.getName() + " where value like $1");
        query.bind(nameUpdated);
        result = (Collection) query.execute();
        assertNotNull(result);
        assertEquals(1, result.size());
    }

    public void testLockBasics() throws Exception
    {
        TransactionImpl tx1 = (TransactionImpl) odmg1.newTransaction();
        TransactionImpl tx2 = (TransactionImpl) odmg2.newTransaction();

        Article a = new Article();
        a.setArticleId(333);

        tx1.begin();
        tx2.begin();
        LockManager lm = LockManagerFactory.getLockManager();
        boolean success1 = lm.writeLock(tx1, a);
        boolean success2 = lm.writeLock(tx2, a);

        boolean success3 = lm.releaseLock(tx1, a);

        assertTrue("1st lock should succeed", success1);
        assertTrue("2nd lock should not succeed", !success2);
        assertTrue("release should succeed", success3);

        try
        {
            tx1.abort();
            tx2.abort();
        }
        catch (Exception e)
        {
        }
    }

    public void testLockBasics2() throws Exception
    {
        TransactionImpl tx1 = (TransactionImpl) odmg1.newTransaction();
        TransactionImpl tx2 = (TransactionImpl) odmg2.newTransaction();

        Article a1 = new Article();
        a1.setArticleId(333);

        Article a2 = new Article();
        a2.setArticleId(333);

        tx1.begin();
        tx2.begin();
        LockManager lm = LockManagerFactory.getLockManager();

        assertFalse(tx1.getGUID().equals(tx2.getGUID()));

        assertTrue("1st lock should succeed", lm.writeLock(tx1, a1));
        assertFalse("2nd lock should not succeed", lm.writeLock(tx2, a2));
        lm.releaseLock(tx2, a2);
        lm.releaseLock(tx2, a1);
        assertFalse(lm.checkWrite(tx2, a1));
        assertFalse(lm.checkWrite(tx2, a2));
        assertTrue(lm.checkWrite(tx1, a1));
        assertTrue(lm.checkWrite(tx1, a2));
        //assertFalse("2nd release should not succeed", lm.releaseLock(tx2, a2));
        //assertFalse("2nd release should not succeed", lm.releaseLock(tx2, a1));
        assertTrue("release should succeed", lm.releaseLock(tx1, a2));
        assertTrue("2nd object lock should succeed", lm.writeLock(tx2, a2));
        assertTrue("release 2nd object lock should succeed", lm.releaseLock(tx2, a2));

        try
        {
            tx1.abort();
            tx2.abort();
        }
        catch (Exception e)
        {
        }
    }

	/**
	 * test proper treatment of Optimistic Locking in
	 * ODMG transactions
	 */
    public void testOptimisticLockBasics() throws Exception
    {
        TransactionImpl tx1 = (TransactionImpl) odmg1.newTransaction();
        TransactionImpl tx2 = (TransactionImpl) odmg2.newTransaction();

        LockedByVersion obj = new LockedByVersion();


        tx1.begin();

        tx1.lock(obj,Transaction.WRITE);
		obj.setValue("tx1");
		tx1.commit();

		obj.setVersion(obj.getVersion() - 1);
		tx2.begin();
		tx2.lock(obj,Transaction.WRITE);

		obj.setValue("tx2");
		boolean signalOLException = false;
		try
		{
			tx2.commit();
		}
		// OL exceptions should be signalled as ODMG LockNotGrantedExceptions
		// so that users can react accordingly
		catch (LockNotGrantedException ex)
		{
			signalOLException = true;
		}
		assertTrue("expected that a OL exception is caught", signalOLException);
    }


    /**
     * factory method that createa an PerformanceArticle
     * @return the created PerformanceArticle object
     */
    private Article createArticle(String name)
    {
        Article a = new Article();

        a.setArticleName(PRE + name);
        a.setMinimumStock(100);
        a.setOrderedUnits(17);
        a.setPrice(0.45);
        a.setProductGroupId(1);
        a.setStock(234);
        a.setSupplierId(4);
        a.setUnit("bottle");
        return a;
    }

    public void testLockLoop() throws Exception
    {
        int loops = 10;
        Article[] arr = new Article[loops];
        for (int i = 0; i < loops; i++)
        {
            Article a = createArticle("testLockLoop");
            arr[i] = a;
        }

        TransactionImpl tx = (TransactionImpl) odmg1.newTransaction();
        tx.begin();
        for (int i = 0; i < arr.length; i++)
        {
            tx.lock(arr[i], Transaction.WRITE);
        }
        LockManager lm = LockManagerFactory.getLockManager();
        boolean success = lm.writeLock(tx, arr[(loops - 2)]);
        assertTrue("lock should succeed", success);
        tx.commit();

        TransactionImpl tx2 = (TransactionImpl) odmg2.newTransaction();
        tx2.begin();
        success = lm.writeLock(tx2, arr[(loops - 2)]);
        assertTrue("lock should succeed", success);

        OQLQuery query = odmg2.newOQLQuery();
        String sql = "select allArticles from " + Article.class.getName() +
                " where articleName = \"" + PRE + "testLockLoop" + "\"";
        query.create(sql);
        int result = ((List) query.execute()).size();
        tx2.commit();
        assertEquals("Wrong number of objects wrote to DB", loops, result);

        PersistenceBroker broker = PersistenceBrokerFactory.defaultPersistenceBroker();
        broker.clearCache();
        Criteria crit = new Criteria();
        crit.addLike("articleName", PRE + "testLockLoop");
        result = broker.getCount(QueryFactory.newQuery(Article.class, crit));
        broker.close();
        assertEquals("Wrong number of objects wrote to DB", loops, result);
    }
}
