package xdoclet.modules.ojb.model;

/* Copyright 2004-2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.util.*;

import xdoclet.modules.ojb.CommaListIterator;

/**
 * Represents the model used for generating the torque database schema.
 */
public class TorqueModelDef extends DefBase
{
    /** The tables keyed by their names*/
    private SortedMap _tableDefs = new TreeMap();

    /**
     * Generates a new torque database model from the given ojb model.
     * 
     * @param dbName   The name of the database
     * @param ojbModel The ojb model
     * @return The torque model
     */
    public TorqueModelDef(String dbName, ModelDef ojbModel)
    {
        super(dbName);

        ClassDescriptorDef classDef;
        TableDef           tableDef;
        FieldDescriptorDef fieldDef;
        ColumnDef          columnDef;
        String             name;

        for (Iterator classIt = ojbModel.getClasses(); classIt.hasNext();)
        {
            classDef = (ClassDescriptorDef)classIt.next();
            if (classDef.getBooleanProperty(PropertyHelper.OJB_PROPERTY_OJB_PERSISTENT, false) &&
                classDef.getBooleanProperty(PropertyHelper.OJB_PROPERTY_GENERATE_TABLE_INFO, true))
            {
                addTableFor(classDef);
            }
        }
    }

    // The conversion algorithm
    // Note that the complete algorithm is here rather than splitting in onto the
    // various torque model elements because maintaining it is easier this way
    
    /**
     * Adds a table for the given class descriptor (if necessary).
     * 
     * @param classDef The class descriptor
     */
    private void addTableFor(ClassDescriptorDef classDef)
    {
        String                  name     = classDef.getProperty(PropertyHelper.OJB_PROPERTY_TABLE);
        TableDef                tableDef = getTable(name);
        FieldDescriptorDef      fieldDef;
        ReferenceDescriptorDef  refDef;
        CollectionDescriptorDef collDef;
        ColumnDef               columnDef;
        IndexDef                indexDef;

        if (tableDef == null)
        {
            tableDef = new TableDef(name);
            addTable(tableDef);
            if (classDef.hasProperty(PropertyHelper.OJB_PROPERTY_DOCUMENTATION))
            {
                tableDef.setProperty(PropertyHelper.OJB_PROPERTY_DOCUMENTATION,
                                     classDef.getProperty(PropertyHelper.OJB_PROPERTY_DOCUMENTATION));
            }
        }
        for (Iterator fieldIt = classDef.getFields(); fieldIt.hasNext();)
        {
            fieldDef = (FieldDescriptorDef)fieldIt.next();
            if (fieldDef.getBooleanProperty(PropertyHelper.OJB_PROPERTY_IGNORE, false))
            {
                continue;
            }
            columnDef = addColumnFor(fieldDef, tableDef);
            if (fieldDef.getBooleanProperty(PropertyHelper.OJB_PROPERTY_INDEXED, false))
            {
                // add the field to the default index
                indexDef = tableDef.getIndex(null);
                if (indexDef == null)
                {
                    indexDef = new IndexDef(null, false);
                    tableDef.addIndex(indexDef);
                }
                indexDef.addColumn(columnDef.getName());
            }
        }
        for (Iterator refIt = classDef.getReferences(); refIt.hasNext();)
        {
            refDef = (ReferenceDescriptorDef)refIt.next();
            if (!refDef.getBooleanProperty(PropertyHelper.OJB_PROPERTY_IGNORE, false))
            {
                addForeignkeys(refDef, tableDef);
            }
        }
        for (Iterator collIt = classDef.getCollections(); collIt.hasNext();)
        {
            collDef = (CollectionDescriptorDef)collIt.next();
            if (!collDef.getBooleanProperty(PropertyHelper.OJB_PROPERTY_IGNORE, false) &&
                collDef.hasProperty(PropertyHelper.OJB_PROPERTY_INDIRECTION_TABLE))
            {
                addIndirectionTable(collDef);
            }
        }
        for (Iterator indexIt = classDef.getIndexDescriptors(); indexIt.hasNext();)
        {
            addIndex((IndexDescriptorDef)indexIt.next(), tableDef);
        }
    }

    /**
     * Generates a column for the given field and adds it to the table.
     * 
     * @param fieldDef The field
     * @param tableDef The table
     * @return The column def
     */
    private ColumnDef addColumnFor(FieldDescriptorDef fieldDef, TableDef tableDef)
    {
        String    name      = fieldDef.getProperty(PropertyHelper.OJB_PROPERTY_COLUMN);
        ColumnDef columnDef = tableDef.getColumn(name);

        if (columnDef == null)
        {
            columnDef = new ColumnDef(name);
            tableDef.addColumn(columnDef);
        }
        if (!fieldDef.isNested())
        {    
            columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_JAVANAME, fieldDef.getName());
        }
        columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_TYPE, fieldDef.getProperty(PropertyHelper.OJB_PROPERTY_JDBC_TYPE));
        columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_ID, fieldDef.getProperty(PropertyHelper.OJB_PROPERTY_ID));
        if (fieldDef.getBooleanProperty(PropertyHelper.OJB_PROPERTY_PRIMARYKEY, false))
        {
            columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_PRIMARYKEY, "true");
            columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_REQUIRED, "true");
        }
        else if (!fieldDef.getBooleanProperty(PropertyHelper.OJB_PROPERTY_NULLABLE, true))
        {
            columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_REQUIRED, "true");
        }
        if ("database".equals(fieldDef.getProperty(PropertyHelper.OJB_PROPERTY_AUTOINCREMENT)))
        {
            columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_AUTOINCREMENT, "true");
        }
        columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_SIZE, fieldDef.getSizeConstraint());
        if (fieldDef.hasProperty(PropertyHelper.OJB_PROPERTY_DOCUMENTATION))
        {
            columnDef.setProperty(PropertyHelper.OJB_PROPERTY_DOCUMENTATION,
                                  fieldDef.getProperty(PropertyHelper.OJB_PROPERTY_DOCUMENTATION));
        }
        return columnDef;
    }

    /**
     * Adds a foreignkey for the reference to the table.
     * 
     * @param refDef   The reference
     * @param tableDef The table
     */
    private void addForeignkeys(ReferenceDescriptorDef refDef, TableDef tableDef)
    {
        // a foreignkey is added to the table schema if
        // the referenced table exists (i.e. the referenced type has an associated table)
        // then the foreignkey consists of:
        //   remote table  = table of referenced type
        //   local fields  = foreignkey fields of the reference 
        //   remote fields = primarykeys of the referenced type
        ClassDescriptorDef ownerClassDef      = (ClassDescriptorDef)refDef.getOwner();
        ClassDescriptorDef referencedClassDef = ((ModelDef)ownerClassDef.getOwner()).getClass(refDef.getProperty(PropertyHelper.OJB_PROPERTY_CLASS_REF));

        if (!referencedClassDef.getBooleanProperty(PropertyHelper.OJB_PROPERTY_GENERATE_TABLE_INFO, true))
        {
            // the referenced type is not persistent
            return;
        }

        try
        {
            String             name          = refDef.getName();
            String             tableName     = referencedClassDef.getProperty(PropertyHelper.OJB_PROPERTY_TABLE);
            ArrayList          localFields   = ownerClassDef.getFields(refDef.getProperty(PropertyHelper.OJB_PROPERTY_FOREIGNKEY));
            ArrayList          remoteFields  = referencedClassDef.getPrimaryKeys();
            ForeignkeyDef      foreignkeyDef = new ForeignkeyDef(name, tableName);
            FieldDescriptorDef localField;
            FieldDescriptorDef remoteField;
            
            // the field arrays have the same length if we already checked the constraints
            for (int idx = 0; idx < localFields.size(); idx++)
            {
                localField  = (FieldDescriptorDef)localFields.get(idx);
                remoteField = (FieldDescriptorDef)remoteFields.get(idx);
                foreignkeyDef.addColumnPair(localField.getProperty(PropertyHelper.OJB_PROPERTY_COLUMN),
                                            remoteField.getProperty(PropertyHelper.OJB_PROPERTY_COLUMN));
            }
            tableDef.addForeignkey(foreignkeyDef);
        }
        catch (NoSuchFieldException ex)
        {
            // won't happen if we already checked the constraints
        }
    }

    /**
     * Adds an index to the table for the given index descriptor.
     * 
     * @param indexDescDef The index descriptor
     * @param tableDef     The table
     */
    private void addIndex(IndexDescriptorDef indexDescDef, TableDef tableDef)
    {
        IndexDef indexDef = tableDef.getIndex(indexDescDef.getName());

        if (indexDef == null)
        {
            indexDef = new IndexDef(indexDescDef.getName(),
                                    indexDescDef.getBooleanProperty(PropertyHelper.OJB_PROPERTY_UNIQUE, false));
            tableDef.addIndex(indexDef);
        }

        try
        {
            String             fieldNames = indexDescDef.getProperty(PropertyHelper.OJB_PROPERTY_FIELDS);
            ArrayList          fields     = ((ClassDescriptorDef)indexDescDef.getOwner()).getFields(fieldNames);
            FieldDescriptorDef fieldDef;

            for (Iterator it = fields.iterator(); it.hasNext();)
            {
                fieldDef = (FieldDescriptorDef)it.next();
                indexDef.addColumn(fieldDef.getProperty(PropertyHelper.OJB_PROPERTY_COLUMN));
            }
        }
        catch (NoSuchFieldException ex)
        {
            // won't happen if we already checked the constraints
        }
    }

    /**
     * Adds the indirection table for the given collection descriptor.
     * 
     * @param collDef The collection descriptor
     */
    private void addIndirectionTable(CollectionDescriptorDef collDef)
    {
        String   tableName = collDef.getProperty(PropertyHelper.OJB_PROPERTY_INDIRECTION_TABLE);
        TableDef tableDef  = getTable(tableName);

        if (tableDef == null)
        {
            tableDef = new TableDef(tableName);
            addTable(tableDef);
        }

        // we add columns for every primarykey in this and the element type
        //   collection.foreignkeys        <-> ownerclass.primarykeys
        //   collection.remote-foreignkeys <-> elementclass.primarykeys
        // we also add foreignkeys to the table
        //   name is empty (default foreignkey)
        //   remote table   = table of ownerclass/elementclass
        //   local columns  = columns in indirection table 
        //   remote columns = columns of corresponding primarykeys in ownerclass/elementclass
        ClassDescriptorDef ownerClassDef    = (ClassDescriptorDef)collDef.getOwner();
        ClassDescriptorDef elementClassDef  = ((ModelDef)ownerClassDef.getOwner()).getClass(collDef.getProperty(PropertyHelper.OJB_PROPERTY_ELEMENT_CLASS_REF));
        ArrayList          localPrimFields  = ownerClassDef.getPrimaryKeys();
        ArrayList          remotePrimFields = elementClassDef.getPrimaryKeys();
        String             ownerTable       = ownerClassDef.getProperty(PropertyHelper.OJB_PROPERTY_TABLE);
        String             elementTable     = elementClassDef.getProperty(PropertyHelper.OJB_PROPERTY_TABLE);
        CommaListIterator  localKeys        = new CommaListIterator(collDef.getProperty(PropertyHelper.OJB_PROPERTY_FOREIGNKEY));
        CommaListIterator  remoteKeys       = new CommaListIterator(collDef.getProperty(PropertyHelper.OJB_PROPERTY_REMOTE_FOREIGNKEY));
        FieldDescriptorDef fieldDef;
        ForeignkeyDef      foreignkey       = null;
        ColumnDef          columnDef;
        String             relationName;
        String             name;
        int                idx;

        relationName = collDef.getProperty(PropertyHelper.TORQUE_PROPERTY_RELATION_NAME);
        if (relationName != null)
        {    
            foreignkey   = tableDef.getForeignkey(relationName, ownerTable);
            if (foreignkey == null)
            {
                foreignkey = new ForeignkeyDef(relationName, ownerTable);
                tableDef.addForeignkey(foreignkey);
            }
        }
        for (idx = 0; localKeys.hasNext(); idx++)
        {
            fieldDef  = (FieldDescriptorDef)localPrimFields.get(idx);
            name      = localKeys.getNext();

            columnDef = tableDef.getColumn(name);
            if (columnDef == null)
            {
                columnDef = new ColumnDef(name);
                tableDef.addColumn(columnDef);
            }
            columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_TYPE, fieldDef.getProperty(PropertyHelper.OJB_PROPERTY_JDBC_TYPE));
            columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_SIZE, fieldDef.getSizeConstraint());

            if (foreignkey != null)
            {    
                foreignkey.addColumnPair(name, fieldDef.getProperty(PropertyHelper.OJB_PROPERTY_COLUMN));
            }
        }

        relationName = collDef.getProperty(PropertyHelper.TORQUE_PROPERTY_INV_RELATION_NAME);
        if (relationName != null)
        {    
            foreignkey   = tableDef.getForeignkey(relationName, elementTable);
            if (foreignkey == null)
            {
                foreignkey = new ForeignkeyDef(relationName, elementTable);
                tableDef.addForeignkey(foreignkey);
            }
        }
        for (idx = 0; remoteKeys.hasNext(); idx++)
        {
            fieldDef  = (FieldDescriptorDef)remotePrimFields.get(idx);
            name      = remoteKeys.getNext();

            columnDef = tableDef.getColumn(name);
            if (columnDef == null)
            {
                columnDef = new ColumnDef(name);
                tableDef.addColumn(columnDef);
            }
            columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_TYPE, fieldDef.getProperty(PropertyHelper.OJB_PROPERTY_JDBC_TYPE));
            columnDef.setProperty(PropertyHelper.TORQUE_PROPERTY_SIZE, fieldDef.getSizeConstraint());

            if (foreignkey != null)
            {    
                foreignkey.addColumnPair(name, fieldDef.getProperty(PropertyHelper.OJB_PROPERTY_COLUMN));
            }
        }
    }

    // Access methods
    
    /**
     * Returns an iterator of the tables.
     * 
     * @return The tables
     */
    public Iterator getTables()
    {
        return _tableDefs.values().iterator();
    }
    
    /**
     * Returns the table of the given name if it exists.
     * 
     * @param name The table name
     * @return The table def or <code>null</code> if there is no such table
     */
    public TableDef getTable(String name)
    {
        return (TableDef)_tableDefs.get(name);
    }

    /**
     * Adds a table to this model.
     * 
     * @param table The table
     */
    private void addTable(TableDef table)
    {
        table.setOwner(this);
        _tableDefs.put(table.getName(), table);
    }
}
